from __future__ import annotations

import copy
import inspect
from operator import attrgetter
from typing import Union, Tuple, List, Any, TypeVar, Mapping, Generic, FrozenSet
from collections.abc import Hashable

from typic.util import cached_property

__all__ = ("FrozenDict", "freeze")

KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT", covariant=True)  # Value type.

_hashgetter = attrgetter("__hash__")


class FrozenDict(Generic[KT, VT], dict):
    """An immutable, hashable mapping.

    This inherits directly from the builtin :py:class:`dict`.

    Notes
    -----
    Because it must recursively freeze all nested items to ensure it is hashable,
    this data-structure is substantially slower on init than a native dict.

    This operation is highly optimized, but the recursive nature is still expensive.

    However, once initialized, it is equivalent in performance to a native dict in every
    other way. The recommended use-case for this object is for compile-time objects
    (such as configurations, dictionary-dispatching, etc.).

    Examples
    --------
    >>> import typic
    >>> fdict = typic.FrozenDict({"foo": ["bar"]})
    >>> typic.ishashable(fdict)
    True
    >>> fdict["foo"]
    ('bar',)
    >>> fdict.update(foo=["car"])
    Traceback (most recent call last):
    ...
    TypeError: attempting to mutate immutable type 'FrozenDict'
    >>> del fdict["foo"]
    Traceback (most recent call last):
    ...
    TypeError: attempting to mutate immutable type 'FrozenDict'
    >>> fdict.pop("foo")
    Traceback (most recent call last):
    ...
    TypeError: attempting to mutate immutable type 'FrozenDict'
    >>> fdict.clear()
    Traceback (most recent call last):
    ...
    TypeError: attempting to mutate immutable type 'FrozenDict'
    """

    _MSG = "attempting to mutate immutable type 'FrozenDict'"

    def __init__(
        self,
        __obj: Union[Mapping, List[Tuple[str, Hashable]]] = None,
        *,
        __hashgetter=_hashgetter,
        **kwargs,
    ):
        super().__init__(
            {
                x: y if __hashgetter(y) else freeze(y)
                for x, y in {**(dict(__obj or {})), **kwargs}.items()
            }
        )

    def __copy__(self) -> "FrozenDict":
        return self.__class__({**self})

    def __deepcopy__(self, memodict: dict = None) -> "FrozenDict":
        return self.__class__({x: copy.deepcopy(y, memodict) for x, y in self.items()})

    @cached_property
    def __hash(self) -> int:
        return hash(frozenset(self.items()))

    def __hash__(self) -> int:  # type: ignore
        return self.__hash

    def __setitem__(self, key, value):
        """Mutations are disallowed."""
        raise TypeError(self._MSG) from None

    def __delitem__(self, key):
        """Mutations are disallowed."""
        raise TypeError(self._MSG) from None

    def pop(self, k):
        """Mutations are disallowed."""
        raise TypeError(self._MSG) from None

    def popitem(self):
        """Mutations are disallowed."""
        raise TypeError(self._MSG) from None

    def clear(self):
        """Mutations are disallowed."""
        raise TypeError(self._MSG) from None

    def update(self, *args, **kwargs):
        """Mutations are disallowed."""
        raise TypeError(self._MSG) from None

    def setdefault(self, *args, **kwargs):
        """Mutations are disallowed."""
        raise TypeError(self._MSG) from None

    def mutate(self, other: Mapping = None, **kwargs) -> "FrozenDict":
        """Return a new :py:class:`FrozenDict` with changes merged in.

        Priority of keys is in inverse order, i.e.:
            1. `**kwargs`
            2. `other` mapping
            3. `self`, the object you're mutating.

        Examples
        --------
        >>> import typic
        >>> fdict = typic.FrozenDict(foo=["bar"])
        >>> fdict.mutate({"bazz": "buzz"}, bazz="blah")
        {'foo': ('bar',), 'bazz': 'blah'}
        """
        return self.__class__({**self, **(other or {}), **kwargs})


FrozenT = Union[FrozenDict, Hashable, Tuple, FrozenSet, None]


def freeze(o: Any, *, __hashgetter=_hashgetter) -> FrozenT:
    if __hashgetter(o) or inspect.isclass(o) or inspect.isfunction(o):
        return o

    if isinstance(o, set):
        return frozenset(o)

    if isinstance(o, Mapping):
        return FrozenDict(
            {x: y if __hashgetter(y) else freeze(y) for x, y in o.items()}
        )

    return (*(x if __hashgetter(x) else freeze(x) for x in o),)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# flake8: noqa
from . import types, constraints
from .checks import *
from .constraints import *
from .ext.schema import *
from .klass import klass, field
from .types import *
from .util import *

# NOTE: This import must come *last*
from .api import *

al = typed



from __future__ import annotations

import dataclasses
from types import MappingProxyType
from typing import (
    Type,
    ClassVar,
    Tuple,
    Sequence,
    Dict,
    List,
    Any,
    Iterator,
    Union,
    Optional,
    Hashable,
    Set,
    FrozenSet,
)

from typic import gen, checks, util
from typic.types.frozendict import freeze
from .common import BaseConstraints, ContextT, AssertionsT, ConstraintsProtocolT

Array = Union[FrozenSet, Set, List, Tuple]
"""The supported builtin types for defining restricted array-types."""


def unique_fast(
    seq: Sequence, *, ret_type: Type[Union[list, tuple]] = list
) -> Sequence:
    """Fastest order-preserving method for (hashable) uniques in Python >= 3.6.

    Notes
    -----
    Values of seq must be hashable!

    See Also
    --------
    `Uniquify List in Python 3.6 <https://www.peterbe.com/plog/fastest-way-to-uniquify-a-list-in-python-3.6>`_
    """
    return ret_type(dict.fromkeys(seq))


def unique_slow(
    seq: Sequence, *, ret_type: Type[Union[list, tuple]] = list
) -> Sequence:
    """Fastest order-preserving method for (unhashable) uniques in Python >= 3.6.

    See Also
    --------
    `Uniquify List in Python 3.6 <https://www.peterbe.com/plog/fastest-way-to-uniquify-a-list-in-python-3.6>`_
    """
    seen: Set[Hashable] = set()
    return ret_type(_unique_slow(seq, seen))


def _get_hash(obj: Any):
    if checks.ishashable(obj):
        return hash(obj)
    if dataclasses.is_dataclass(obj):
        obj = dataclasses.asdict(obj)
    return hash(freeze(obj))


def _unique_slow(seq: Sequence, seen: set) -> Iterator[Any]:
    add = seen.add
    for x in seq:
        h = _get_hash(x)
        if h in seen:
            continue
        add(h)
        yield x


def unique(seq: Sequence, *, ret_type: Type[Union[list, tuple]] = list) -> Sequence:
    """Fastest, order-preserving method for uniques in Python >= 3.6.

    See Also
    --------
    `Uniquify List in Python 3.6 <https://www.peterbe.com/plog/fastest-way-to-uniquify-a-list-in-python-3.6>`_
    """
    try:
        return unique_fast(seq, ret_type=ret_type)
    except TypeError:
        return unique_slow(seq, ret_type=ret_type)


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class ArrayConstraints(BaseConstraints):
    """Specific constraints pertaining to a sized, array-like type.

    These constraints are meant to align closely to JSON Schema type constraints.

    Notes
    -----
    Doesn't support mappings, but could be updated to do so.
    """

    type: ClassVar[Type[Array]]
    """The type of array the input should be."""
    min_items: Optional[int] = None
    """The minimum number of items which must be present in the array."""
    max_items: Optional[int] = None
    """The maximum number of items which may be present in the array."""
    unique: Optional[bool] = None
    """Whether this array should only have unique items.

    Notes
    -----
    Rather than reject arrays which are not unique, we will simply make the array unique.
    """
    values: Optional[ConstraintsProtocolT] = None
    """The constraints for which the items in the array must adhere.

    This can be a single type-constraint, or a tuple of multiple constraints.
    """

    def _get_assertions(self) -> AssertionsT:
        asserts: List[str] = []
        if self.min_items is not None:
            asserts.append(f"size >= {self.min_items}")
        if self.max_items is not None:
            asserts.append(f"size <= {self.max_items}")
        return asserts

    def _build_assertions(self, func: gen.Block, assertions: AssertionsT):
        # Only get the size if we have to.
        if assertions:
            if (self.max_items, self.min_items) != (None, None):
                func.l(f"size = len({self.VALUE})")
            BaseConstraints._build_assertions(self, func=func, assertions=assertions)

    def _build_validator(
        self, func: gen.Block, context: ContextT, assertions: AssertionsT
    ) -> ContextT:
        # If we don't have a natively unique type and we're supposed to be unique, make it so.
        if self.unique is True and util.origin(self.type) not in {set, frozenset}:
            func.l(f"{self.VALUE} = unique({self.VALUE})", unique=unique)
        context = BaseConstraints._build_validator(self, func, context, assertions)
        # Validate the items if necessary.
        if self.values:
            o = util.origin(self.type)
            itval = "__item_validator"
            ctx = {
                "unique": unique,
                itval: self.values.validate,
                o.__name__: o,
                "_lazy_repr": util.collectionrepr,
            }
            r = "i" if issubclass(self.type, Sequence) else "x"
            field = f"_lazy_repr({self.FNAME}, {r})"
            func.l(
                f"{self.VALUE} = "
                f"{o.__name__}("
                f"({itval}(x, field={field}) for i, x in enumerate({self.VALUE}))"
                f")",
                **ctx,  # type: ignore
            )
        return context

    def for_schema(self, *, with_type: bool = False) -> dict:
        schema: Dict[str, Any] = dict(
            title=self.name,
            minItems=self.min_items,
            maxItems=self.max_items,
            uniqueItems=self.unique,
        )
        if with_type:
            schema["type"] = "array"
        return {x: y for x, y in schema.items() if y is not None}


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class ListConstraints(ArrayConstraints):
    """Specific constraints pertaining to a :py:class:`list`."""

    type: ClassVar[Type[list]] = list


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class TupleConstraints(ArrayConstraints):
    """Specific constraints pertaining to a :py:class:`tuple`."""

    type: ClassVar[Type[tuple]] = tuple
    values: Union[Optional[ConstraintsProtocolT], Sequence[ConstraintsProtocolT]] = None  # type: ignore

    def _build_validator(
        self, func: gen.Block, context: ContextT, assertions: AssertionsT
    ) -> ContextT:
        if isinstance(self.values, Sequence):
            if self.unique is True:
                func.l(
                    f"{self.VALUE} = unique({self.VALUE}, ret_type=tuple)",
                    unique=unique,
                )
            item_validators = MappingProxyType(
                {i: c.validate for i, c in enumerate(self.values)}
            )
            o = util.origin(self.type)
            itval = "__item_validators"
            ctx = {
                "unique": unique,
                itval: item_validators,
                o.__name__: o,
                "_lazy_repr": util.collectionrepr,
            }
            field = f"_lazy_repr({self.FNAME}, i)"
            func.l(
                f"{self.VALUE} = "
                f"{o.__name__}("
                f"({itval}[i](x, field={field}) if i in {itval} else x "
                f"for i, x in enumerate({self.VALUE}))"
                f")",
                **ctx,  # type: ignore
            )
            return ctx
        return ArrayConstraints._build_validator(
            self, func=func, context=context, assertions=assertions
        )


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class SetContraints(ArrayConstraints):
    """Specific constraints pertaining to a :py:class:`set`."""

    type: ClassVar[Type[set]] = set
    unique: bool = True


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class FrozenSetConstraints(ArrayConstraints):
    """Specific constraints pertaining to a :py:class:`frozenset`."""

    type: ClassVar[Type[frozenset]] = frozenset
    unique: bool = True



from __future__ import annotations

import dataclasses
import decimal
from typing import Union, Type, ClassVar, Optional, Dict, List

from typic import gen, util
from .common import BaseConstraints, ContextT, AssertionsT
from .error import ConstraintSyntaxError, ConstraintValueError

NumberT = Union[int, float, decimal.Decimal]


def _get_digits(tup: decimal.DecimalTuple):
    if tup.exponent >= 0:
        # A positive exponent adds that many trailing zeros.
        digits = len(tup.digits) + tup.exponent
        decimals = 0
    else:
        # If the absolute value of the negative exponent is larger than the
        # number of digits, then it's the same as the number of digits,
        # because it'll consume all of the digits in digit_tuple and then
        # add abs(exponent) - len(digit_tuple) leading zeros after the
        # decimal point.
        if abs(tup.exponent) > len(tup.digits):
            digits = decimals = abs(tup.exponent)
        else:
            digits = len(tup.digits)
            decimals = abs(tup.exponent)
    whole_digits = digits - decimals
    return whole_digits, digits, decimals


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class NumberConstraints(BaseConstraints):
    """Specific constraints pertaining to number-like types.

    Currently supports :py:class:`int`, :py:class:`float`, and
    :py:class:`decimal.Decimal`.

    See Also
    --------
    :py:class:`~typic.types.constraints.common.BaseConstraints`
    """

    type: ClassVar[Type[NumberT]]
    """The builtin type for this constraint."""
    gt: Optional[NumberT] = None
    """The value inputs must be greater-than."""
    ge: Optional[NumberT] = None
    """The value inputs must be greater-than-or-equal-to."""
    lt: Optional[NumberT] = None
    """The value inputs must be less-than."""
    le: Optional[NumberT] = None
    """The value inputs must be less-than-or-equal-to."""
    mul: Optional[NumberT] = None
    """The value inputs must be a multiple-of."""

    def _check_syntax(self):
        if self.gt is not None:
            if self.ge is not None:
                raise ConstraintSyntaxError(
                    "Values must either be '>' or '>=', not both."
                ) from None
            if self.gt in {self.lt, self.le}:
                msg = (
                    f"Values for '>' and '<|<=' are equal ({self.gt}, {self.lt}|{self.le}), "
                    f"this will always be false."
                )
                raise ConstraintSyntaxError(msg) from None
        if self.lt is not None:
            if self.le is not None:
                raise ConstraintSyntaxError(
                    "Values must either be '<' or '<=', not both."
                ) from None
            if self.lt in {self.gt, self.ge}:
                msg = (
                    f"Values for '<' and '>|>=' are equal ({self.lt}, {self.gt}|{self.ge}), "
                    f"this will always be false."
                )
                raise ConstraintSyntaxError(msg) from None

    def _get_assertions(self) -> AssertionsT:
        asserts: List[str] = []
        if self.gt is not None:
            asserts.append(f"{self.VALUE} > {self.gt}")
        if self.ge is not None:
            asserts.append(f"{self.VALUE} >= {self.ge}")
        if self.lt is not None:
            asserts.append(f"{self.VALUE} < {self.lt}")
        if self.le is not None:
            asserts.append(f"{self.VALUE} <= {self.le}")
        if self.mul is not None:
            asserts.append(f"{self.VALUE} % {self.mul} == 0")
        return asserts

    def for_schema(self, *, with_type: bool = False) -> dict:
        schema: Dict[str, Union[None, NumberT, str]] = dict(
            title=self.name,
            multipleOf=self.mul,
            minimum=self.ge,
            maximum=self.le,
            exclusiveMinimum=self.gt,
            exclusiveMaximum=self.lt,
        )
        if with_type:
            schema["type"] = "number"
        return {x: y for x, y in schema.items() if y is not None}


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class IntContraints(NumberConstraints):
    """Constraints specifically for :py:class:`int`.

    See Also
    --------
    :py:class:`NumberConstraints`
    """

    type: ClassVar[Type[NumberT]] = int

    def for_schema(self, *, with_type: bool = False) -> dict:
        schema = NumberConstraints.for_schema(self)
        if with_type:
            schema["type"] = "integer"
        return schema


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class FloatContraints(NumberConstraints):
    """Constraints specifically for :py:class:`int`.

    See Also
    --------
    :py:class:`NumberConstraints`
    """

    type: ClassVar[Type[NumberT]] = float


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class DecimalContraints(NumberConstraints):
    """Constraints specifically for :py:class:`int`.

    See Also
    --------
    :py:class:`NumberConstraints`
    """

    type: ClassVar[Type[NumberT]] = decimal.Decimal
    max_digits: Optional[int] = None
    """The maximum allowed digits for the input."""
    decimal_places: Optional[int] = None
    """The maximum allowed decimal places for the input."""

    def _check_syntax(self):
        NumberConstraints._check_syntax(self)
        if None in {self.max_digits, self.decimal_places}:
            return
        if self.max_digits < self.decimal_places:
            msg = (
                f"Contraint <decimal_places={self.decimal_places!r}> "
                f"should never be greater than "
                f"Constraint <max_digits={self.max_digits!r}>"
            )
            raise ConstraintSyntaxError(msg) from None

    def _get_assertions(self) -> AssertionsT:
        asserts = NumberConstraints._get_assertions(self)
        if (self.max_digits, self.decimal_places) == (None, None):
            return asserts

        if self.max_digits is not None:
            asserts.append(f"digits <= {self.max_digits}")
        if self.decimal_places is not None:
            asserts.append(f"decimals <= {self.decimal_places}")
        if self.decimal_places is not None and self.max_digits is not None:
            asserts.append(f"whole <= ({self.max_digits} - {self.decimal_places})")
        return asserts

    def _build_validator(
        self, func: gen.Block, context: ContextT, assertions: AssertionsT
    ) -> ContextT:
        if (self.max_digits, self.decimal_places) == (None, None):
            context = NumberConstraints._build_validator(
                self, func, context=context, assertions=assertions
            )
            context.update(
                decimal=decimal, Decimal=decimal.Decimal, _get_digits=_get_digits
            )
            return context
        # Update the global namespace for the validator
        # Add setup/sanity checks for decimals.
        func.l(f"{self.VALUE} = decimal.Decimal({self.VALUE})")
        with func.b(
            f"if {self.VALUE}.is_infinite():",
            ConstraintValueError=ConstraintValueError,
        ) as b:
            b.l("raise ConstraintValueError('Cannot validate infinite values.')")
        func.l(f"tup = {self.VALUE}.as_tuple()")
        func.l(
            "whole, digits, decimals = _get_digits(tup)",
            _get_digits=_get_digits,
        )
        context = NumberConstraints._build_validator(
            self, func, context=context, assertions=assertions
        )
        context.update(
            decimal=decimal, Decimal=decimal.Decimal, _get_digits=_get_digits
        )
        return context



from __future__ import annotations

import copy
import dataclasses
import datetime
import decimal
import enum
import ipaddress
import pathlib
import re
import reprlib
import uuid
from typing import (
    ClassVar,
    Collection,
    List,
    Mapping,
    Any,
    Union,
    Callable,
    Tuple,
    Type,
    Pattern,
    AnyStr,
    Optional,
    Text,
)

import pendulum

from typic.compat import Literal
from typic.ext.json import dumps
from typic.serde.common import SerdeFlags
from typic.serde.resolver import resolver
from typic.util import filtered_repr, cached_property, TypeMap, ReprT, slotted
from typic.types import dsn, email, frozendict, path, secret, url
from .compat import fastjsonschema

__all__ = (
    "ArraySchemaField",
    "BaseSchemaField",
    "BooleanSchemaField",
    "IntSchemaField",
    "MultiSchemaField",
    "NumberSchemaField",
    "NullSchemaField",
    "ObjectSchemaField",
    "Ref",
    "SchemaFieldT",
    "SchemaType",
    "StringFormat",
    "StrSchemaField",
    "UndeclaredSchemaField",
    "get_field_type",
)


class SchemaType(str, enum.Enum):
    """The official primitive types supported by JSON Schema.

    See Also
    --------
    `JSON Schema Types <https://json-schema.org/understanding-json-schema/reference/type.html>`_
    """

    STR = "string"
    INT = "integer"
    NUM = "number"
    OBJ = "object"
    ARR = "array"
    BOOL = "boolean"
    NULL = "null"


class _Serializable:
    __slots__ = ()

    def primitive(self, *, lazy: bool = False, name: ReprT = None) -> Mapping[str, Any]:
        return resolver.primitive(self, lazy=lazy, name=name)  # type: ignore

    def tojson(self, *, indent: int = 0, ensure_ascii: bool = False, **kwargs) -> str:
        return dumps(
            self.primitive(),
            indent=indent,
            ensure_ascii=ensure_ascii,
            **kwargs,
        )


@slotted
@dataclasses.dataclass(frozen=True)
class Ref(_Serializable):
    """A JSON Schema ref (pointer).

    Usually for directing a validator to another schema definition.
    """

    __serde_flags__ = SerdeFlags(fields={"ref": "$ref"})

    ref: str

    @cached_property
    def title(self) -> str:
        return self.ref.rsplit("/", maxsplit=1)[-1]


class StringFormat(str, enum.Enum):
    """The official string 'formats' supported by JSON Schema.

    See Also
    --------
    `JSON Schema Strings <https://json-schema.org/understanding-json-schema/reference/string.html>`_
    """

    TIME = "time"
    DATE = "date"
    DTIME = "date-time"
    HNAME = "hostname"
    URI = "uri"
    EMAIL = "email"
    UUID = "uuid"
    RE = "regex"
    IPV4 = "ipv4"
    IPV6 = "ipv6"


@slotted(dict=True)
@dataclasses.dataclass(frozen=True, repr=False)
class BaseSchemaField(_Serializable):
    """The base JSON Schema Field."""

    __serde_flags__ = SerdeFlags(omit=(None, NotImplemented))

    type: ClassVar[SchemaType] = NotImplemented
    format: Optional[str] = None
    enum: Optional[Tuple[Any, ...]] = None
    title: Optional[str] = None
    description: Optional[str] = None
    default: Optional[Any] = None
    examples: Optional[List[Any]] = None
    readOnly: Optional[bool] = None
    writeOnly: Optional[bool] = None
    extensions: Optional[Tuple[frozendict.FrozenDict[str, Any], ...]] = None

    __repr = cached_property(reprlib.recursive_repr()(filtered_repr))

    def __repr__(self) -> str:  # pragma: nocover
        return self.__repr

    @cached_property
    def __str(self) -> str:  # pragma: nocover
        fields = (
            [f"type={self.type.value!r}"] if isinstance(self.type, SchemaType) else []
        )
        for f in dataclasses.fields(self):
            val = getattr(self, f.name)
            if (val or val in {False, 0}) and f.repr:
                fields.append(f"{f.name}={val!r}")
        return f"({', '.join(fields)})"

    def __str__(self) -> str:  # pragma: nocover
        return self.__str

    @cached_property
    def validator(self) -> Callable:  # pragma: nocover
        """The JSON Schema validator.

        Notes
        -----
        If `fastjsonschema` is not installed, this will raise a ValueError.

        See Also
        --------
        `fastjsonschema <https://horejsek.github.io/python-fastjsonschema/>`_
        """
        if fastjsonschema:
            return fastjsonschema.compile(self.primitive())
        raise RuntimeError(
            "Can't compile validator, 'fastjsonschema' is not installed."
        )

    def validate(self, obj) -> Any:  # pragma: nocover
        """Validate an object against the defined JSON Schema."""
        try:
            return self.validator(obj)
        except (
            fastjsonschema.JsonSchemaException,
            fastjsonschema.JsonSchemaDefinitionException,
        ):
            raise ValueError(f"<{obj!r}> violates schema: {str(self)}") from None

    def copy(self):  # pragma: nocover
        """Return a new copy of this schema field."""
        return copy.deepcopy(self)


@slotted
@dataclasses.dataclass(frozen=True, repr=False)
class UndeclaredSchemaField(BaseSchemaField):
    """A sentinel object for generating an empty schema."""

    pass


@slotted
@dataclasses.dataclass(frozen=True, repr=False)
class MultiSchemaField(BaseSchemaField):
    """A schema field which supports multiple types."""

    anyOf: Optional[Tuple["SchemaFieldT", ...]] = None
    allOf: Optional[Tuple["SchemaFieldT", ...]] = None
    oneOf: Optional[Tuple["SchemaFieldT", ...]] = None


@slotted
@dataclasses.dataclass(frozen=True, repr=False)
class NullSchemaField(BaseSchemaField):
    type = SchemaType.NULL


@slotted
@dataclasses.dataclass(frozen=True, repr=False)
class StrSchemaField(BaseSchemaField):
    """A JSON Schema Field for the `string` type.

    See Also
    --------
    `JSON Schema String Type <https://json-schema.org/understanding-json-schema/reference/string.html>`_
    `JSON Schema RegEx <https://json-schema.org/understanding-json-schema/reference/regular_expressions.html>`_
    """

    type = SchemaType.STR
    format: Optional[StringFormat] = None
    pattern: Optional[Pattern] = None
    minLength: Optional[int] = None
    maxLength: Optional[int] = None


Number = Union[int, float, decimal.Decimal]


@slotted
@dataclasses.dataclass(frozen=True, repr=False)
class IntSchemaField(BaseSchemaField):
    """A JSON Schema Field for the `integer` type.

    See Also
    --------
    `JSON Schema Numeric Types <https://json-schema.org/understanding-json-schema/reference/numeric.html>`_
    """

    type = SchemaType.INT
    multipleOf: Optional[Number] = None
    maximum: Optional[Number] = None
    minimum: Optional[Number] = None
    exclusiveMaximum: Optional[Number] = None
    exclusiveMinimum: Optional[Number] = None


@slotted
@dataclasses.dataclass(frozen=True, repr=False)
class NumberSchemaField(IntSchemaField):
    """A JSON Schema Field for the `number` type.

    See Also
    --------
    `JSON Schema Numeric Types <https://json-schema.org/understanding-json-schema/reference/numeric.html>`_
    """

    type = SchemaType.NUM


@slotted
@dataclasses.dataclass(frozen=True, repr=False)
class BooleanSchemaField(BaseSchemaField):
    """A JSON Schema Field for the `boolean` type.

    See Also
    --------
    `JSON Schema Boolean Type <https://json-schema.org/understanding-json-schema/reference/boolean.html>`_
    """

    type = SchemaType.BOOL


@slotted
@dataclasses.dataclass(frozen=True, repr=False)
class ObjectSchemaField(BaseSchemaField):
    """A JSON Schema Field for the `object` type.

    See Also
    --------
    `JSON Schema Object Type <https://json-schema.org/understanding-json-schema/reference/object.html>`_
    """

    type = SchemaType.OBJ
    properties: Optional[Mapping[str, Any]] = None
    additionalProperties: Optional[Union[bool, Any]] = None
    maxProperties: Optional[int] = None
    minProperties: Optional[int] = None
    required: Optional[Collection] = None
    propertyNames: Optional[Mapping[str, Pattern]] = None
    patternProperties: Optional[Mapping[Pattern, Any]] = None
    dependencies: Optional[Mapping[str, Union[Tuple[str], Any]]] = None
    definitions: Optional[frozendict.FrozenDict[str, Any]] = None


@slotted
@dataclasses.dataclass(frozen=True, repr=False)
class ArraySchemaField(BaseSchemaField):
    """A JSON Schema Field for the `array` type.

    See Also
    --------
    `JSON Schema Array Type <https://json-schema.org/understanding-json-schema/reference/array.html>`_
    """

    type = SchemaType.ARR
    items: Optional[Any] = None
    contains: Optional[Any] = None
    additionalItems: Optional[bool] = None
    minItems: Optional[int] = None
    maxItems: Optional[int] = None
    uniqueItems: Optional[bool] = None


SchemaFieldT = Union[
    BaseSchemaField,
    StrSchemaField,
    IntSchemaField,
    NumberSchemaField,
    BooleanSchemaField,
    ObjectSchemaField,
    ArraySchemaField,
    MultiSchemaField,
    UndeclaredSchemaField,
    NullSchemaField,
    Ref,
]
"""A type-alias for the defined JSON Schema Fields."""


TYPE_TO_FIELD: Mapping[SchemaType, Type[SchemaFieldT]] = {
    SchemaType.ARR: ArraySchemaField,
    SchemaType.BOOL: BooleanSchemaField,
    SchemaType.INT: IntSchemaField,
    SchemaType.NUM: NumberSchemaField,
    SchemaType.OBJ: ObjectSchemaField,
    SchemaType.STR: StrSchemaField,
}


def get_field_type(type: Optional[Union[SchemaType, Any]]) -> Type[SchemaFieldT]:
    if type is None:
        return MultiSchemaField
    if type is NotImplemented:
        return UndeclaredSchemaField
    return TYPE_TO_FIELD[type]


SCHEMA_FIELD_FORMATS = TypeMap(
    {
        frozendict.FrozenDict: ObjectSchemaField(),
        decimal.Decimal: NumberSchemaField(),
        datetime.datetime: StrSchemaField(format=StringFormat.DTIME),
        pendulum.DateTime: StrSchemaField(format=StringFormat.DTIME),
        datetime.date: StrSchemaField(format=StringFormat.DATE),
        pendulum.Date: StrSchemaField(format=StringFormat.DATE),
        datetime.time: StrSchemaField(format=StringFormat.TIME),
        url.URL: StrSchemaField(format=StringFormat.URI),
        url.AbsoluteURL: StrSchemaField(format=StringFormat.URI),
        url.RelativeURL: StrSchemaField(format=StringFormat.URI),
        dsn.DSN: StrSchemaField(format=StringFormat.URI),
        pathlib.Path: StrSchemaField(format=StringFormat.URI),
        path.FilePath: StrSchemaField(format=StringFormat.URI),
        path.DirectoryPath: StrSchemaField(format=StringFormat.URI),
        path.PathType: StrSchemaField(format=StringFormat.URI),
        url.HostName: StrSchemaField(format=StringFormat.HNAME),
        email.Email: StrSchemaField(format=StringFormat.EMAIL),
        secret.SecretStr: StrSchemaField(),
        secret.SecretBytes: StrSchemaField(),
        uuid.UUID: StrSchemaField(format=StringFormat.UUID),
        re.Pattern: StrSchemaField(format=StringFormat.RE),  # type: ignore
        ipaddress.IPv4Address: StrSchemaField(format=StringFormat.IPV4),
        ipaddress.IPv6Address: StrSchemaField(format=StringFormat.IPV6),
        str: StrSchemaField(),
        AnyStr: StrSchemaField(),  # type: ignore
        Text: StrSchemaField(),
        bytes: StrSchemaField(),
        bool: BooleanSchemaField(),
        int: IntSchemaField(),
        float: NumberSchemaField(),
        list: ArraySchemaField(),
        set: ArraySchemaField(uniqueItems=True),
        tuple: ArraySchemaField(additionalItems=False),
        frozenset: ArraySchemaField(uniqueItems=True, additionalItems=False),
        dict: ObjectSchemaField(),
        type(None): NullSchemaField(),
        Literal: BaseSchemaField(),  # type: ignore
    }
)



from __future__ import annotations

import dataclasses
from types import MappingProxyType
from typing import ClassVar, Pattern, Dict, Set, Match, Optional, Mapping, List
from urllib.parse import urlencode, urlparse, ParseResult, quote, parse_qs

from typic.util import cached_property, slotted
from .secret import SecretStr
from .url import (
    DEFAULT_PORTS,
    INTERNAL_HOSTS,
    INTERNAL_IP_PATTERN,
    NetworkAddress,
    NET_ADDR_PATTERN,
    NetworkAddressValueError,
    PRIVATE_HOSTS,
)

__all__ = ("DSN", "DSNInfo", "DSNValueError")


class DSNValueError(NetworkAddressValueError):
    """A generic error indicating the given value is not a valid DSN.

    Inherits from :py:class:`~typic.types.NetworkAddressValueError`.
    """

    pass


@slotted(dict=True)
@dataclasses.dataclass(frozen=True)
class DSNInfo:
    """Detailed information about a D(ata)S(ource)N(ame).

    Can be called directly, generated by casting a :py:class:`str` as :py:class:`DSN`,
    or created with :py:meth:`DSNInfo.from_str`

    Notes
    -----
    DSNs are *technically* a type of network address, but are more strict.
    There's also a semi-standard API for interacting with them, thanks to SQLAlchemy, etc.
    So we have our own `-Info` object which conforms more closely with the expected API.

    See Also
    --------
    :py:class:`typic.types.NetAddrInfo`
    """

    driver: str
    """The database driver, e.g., `mysql`."""
    username: str
    """The username used for authentication."""
    password: SecretStr
    """The password used for authentication."""
    host: str
    """The host address where the server is located."""
    port: int
    """The exposed port to connect to."""
    name: str
    """The database name."""
    qs: str
    """The query-string to pass on to the database."""
    is_ip: bool = False

    PATTERN: ClassVar[Pattern] = NET_ADDR_PATTERN
    DEFAULT_PORTS: ClassVar[Dict] = DEFAULT_PORTS
    PRIVATE_HOSTS: ClassVar[Set[str]] = PRIVATE_HOSTS
    INTERNAL_HOSTS: ClassVar[Set[str]] = INTERNAL_HOSTS

    @classmethod
    def from_str(cls, value) -> "DSNInfo":
        """Parse & validate a string and generate an instance of :py:class:`DSNInfo`."""
        match: Optional[Match] = cls.PATTERN.match(value)
        if not match or not value:
            raise DSNValueError(f"{value!r} is not a valid DSN.")
        scheme, host = match["scheme"] or "", match["host"] or ""
        if not scheme or not host:
            raise DSNValueError(f"{value!r} is not a valid DSN, missing driver|host.")

        port = int(match["port"] or 0)
        if port == 0 and cls.DEFAULT_PORTS[scheme]:
            port = cls.DEFAULT_PORTS[scheme].copy().pop()
        if port == 0:
            raise DSNValueError(
                f"{value!r} is not a valid DSN, couldn't determine port."
            )
        parsed: ParseResult = urlparse(match["relative"] or "")
        return cls(
            driver=scheme,
            host=host,
            username=match["username"] or "",
            password=SecretStr(match["password"] or ""),
            qs=parsed.query,
            port=port,
            name=parsed.path,
            is_ip=bool(match["ipv4"] or match["ipv6"]),
        )  # type: ignore

    @cached_property
    def base(self) -> str:
        """The 'base' of this DSN.

        Includes driver, user/pass, host, & port"""
        return f"{self.driver}://{self.username}:{self.password.secret}@{self.host}:{self.port}"

    @cached_property
    def relative(self):
        """The 'relative' portion of this DSN.

        Includes the database/schema name and query-string.
        """
        return f"{self.name}{urlencode(self.query)}"

    @cached_property
    def address(self) -> str:
        """The fully-qualified address.

        If this instance was generated from a string, it will match the string exactly,
        EXCEPT if a port wasn't provided in the original string. In those cases, we try
        to determine the appropriate port based upon the driver name provided.
        """
        return f"{self.base}{self.relative}"

    @cached_property
    def address_encoded(self) -> str:
        """The fully-qualified address, encoded."""
        return quote(self.address)  # type: ignore

    @cached_property
    def query(self) -> Mapping[str, List[str]]:
        """The query-string, parsed into a mapping of key -> [values, ...]."""
        return MappingProxyType(parse_qs(self.qs) if self.qs else {})

    @cached_property
    def is_default_port(self) -> bool:
        """Whether or not the port is the default for the SQL dialect."""
        defaults = DEFAULT_PORTS[self.driver.split("+")[0]] | {0}
        return self.port in defaults

    @cached_property
    def is_private(self) -> bool:
        """Whether the host is a private host, i.e., 'localhost'."""
        return self.host in PRIVATE_HOSTS

    @cached_property
    def is_internal(self) -> bool:
        """Whether the host provided is an internal IP/DNS.

        Internal IP/DNS addresses aren't necessarily private, hence the distinction.
        """
        return bool(
            self.host in INTERNAL_HOSTS
            or (self.is_ip and INTERNAL_IP_PATTERN.match(self.host))
        )


# Deepcopy is broken for frozen dataclasses with slots.
# https://github.com/python/cpython/pull/17254
# DSNInfo.__slots__ = tuple(_.name for _ in dataclasses.fields(DSNInfo))


class DSN(NetworkAddress):
    """A Data Source Name string.

    Detailed information about the DSN string can be looked up via :py:attr:`DSN.info`.

    See Also
    --------
    :py:class:`~typic.types.NetworkAddress`

    Examples
    --------
    >>> import typic
    >>> dsn = typic.DSN("postgresql://user:secret@localhost:5432/mydb")
    >>> dsn
    'postgresql://user:secret@localhost:5432/mydb'
    >>> dsn.info.host
    'localhost'
    >>> dsn.info.is_private
    True
    >>> dsn.info.is_default_port
    True
    >>> dsn.info.username
    'user'
    >>> dsn.info.password  # This has been converted to a secret :)
    ******
    >>> dsn.info.name
    '/mydb'
    >>> dsn.info.driver
    'postgresql'
    >>> import json
    >>> json.dumps([dsn])
    '["postgresql://user:secret@localhost:5432/mydb"]'

    Notes
    -----
    This object inherits from :py:class:`str` and, so is natively JSON-serializable.
    """

    @cached_property
    def info(self) -> DSNInfo:  # type: ignore
        """Get detailed information about your DSN string.

        See Also
        --------
        :py:class:`DSNInfo`
        """
        return DSNInfo.from_str(self)



from __future__ import annotations

from typing import Union, Generic, TypeVar


class _Strict:
    __STRICT: bool = False

    def __bool__(self) -> bool:
        return self.__STRICT

    def __str__(self):  # pragma: nocover
        return str(self.__STRICT)

    def __repr__(self):  # pragma: nocover
        return f"<strict-mode={self.__STRICT}>"

    def __hash__(self):
        return hash(self.__STRICT)

    def is_strict_mode(self) -> bool:
        return self.__STRICT

    def strict_mode(self) -> bool:
        """Turn on global ``strict`` mode.

        All resolved annotations will validate their inputs against the generated
        constraints. In some cases, coercion may still be used as the method for
        validation. Additionally, post-validation coercion will occur for
        user-defined classes if needed.

        Notes
        -----
        Global state is messy, but this is provided for convenience. Care must be taken
        when manipulating global state in this way. If you intend to turn on global
        ``strict`` mode, it should be done once, at the start of the application
        runtime, before all annotations have been resolved.

        You cannot toggle ``strict`` mode off once it is enabled during the runtime
        of an application. This is intentional, to limit the potential for hazy or
        unclear state.

        If you find yourself in a situation where you need ``strict`` mode for some
        cases, but not others, you're encouraged to flag ``strict=True`` on the
        decorated class/callable, or even make use of the
        :py:class:`~typic.api.Strict` annotation to flag ``strict`` mode on
        individual fields.
        """
        self.__STRICT = True
        return self.__STRICT

    def _unstrict_mode(self) -> bool:
        self.__STRICT = False
        return self.__STRICT


_T = TypeVar("_T")


class Strict(Generic[_T]):
    pass


StrictStrT = Strict[str]
StrictModeT = Union[bool, _Strict]
STRICT_MODE = _Strict()
is_strict_mode = STRICT_MODE.is_strict_mode
strict_mode = STRICT_MODE.strict_mode



from __future__ import annotations

import dataclasses
from typing import ClassVar, Type, Pattern, Optional, Union, Text

from typic import gen, util
from .common import BaseConstraints, ContextT, AssertionsT


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class TextConstraints(BaseConstraints):
    """Specific constraints pertaining to text-like types (`AnyStr` in Python).

    Currently supports :py:class:`str` and :py:class:`bytes`.

    See Also
    --------
    :py:class:`~typic.types.constraints.common.Constraints`
    """

    type: ClassVar[Type[Union[str, bytes]]]
    """The supported text-types."""
    strip_whitespace: Optional[bool] = None
    """Whether to strip any whitespace from the input."""
    min_length: Optional[int] = None
    """The minimun length this input text must be."""
    max_length: Optional[int] = None
    """The maximum length this input text may be."""
    curtail_length: Optional[int] = None
    """Whether to cut off characters after the defined length."""
    regex: Optional[Pattern[Text]] = None
    """A regex pattern which the input must match."""

    def _get_assertions(self) -> AssertionsT:
        asserts = []
        if self.max_length is not None:
            asserts.append(f"size <= {self.max_length}")
        if self.min_length is not None:
            asserts.append(f"size >= {self.min_length}")
        if self.regex is not None:
            asserts.append(f"__pattern.match({self.VALUE})")
        return asserts

    def _build_validator(
        self, func: gen.Block, context: ContextT, assertions: AssertionsT
    ) -> ContextT:
        # Set up the local env.
        if self.curtail_length is not None:
            func.l(f"{self.VALUE} = {self.VALUE}[:{self.curtail_length}]")
        if self.strip_whitespace:
            func.l(f"{self.VALUE} = {self.VALUE}.strip()")
        if {self.min_length, self.max_length} != {None, None}:
            func.l(f"size = len({self.VALUE})")
        BaseConstraints._build_validator(
            self, func, context=context, assertions=assertions
        )
        # Build the validation.
        if self.regex is not None:
            context.update(__pattern=self.regex)

        return context

    def for_schema(self, *, with_type: bool = False) -> dict:
        schema = dict(
            title=self.name,
            minLength=self.min_length,
            maxLength=self.max_length,
            pattern=self.regex.pattern if self.regex else None,
        )
        if with_type:
            schema["type"] = "string"
        return {x: y for x, y in schema.items() if x is not None}


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class StrConstraints(TextConstraints):
    """Constraints specifically for :py:class:`str`."""

    type: ClassVar[Type[str]] = str


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class BytesConstraints(TextConstraints):
    """Constraints specifically for :py:class:`bytes`."""

    type: ClassVar[Type[bytes]] = bytes



from __future__ import annotations

import abc
import collections.abc
import dataclasses
import enum
import reprlib
import sys
import warnings
from inspect import Signature
from typing import (  # type: ignore
    Callable,
    Any,
    ClassVar,
    Type,
    Tuple,
    List,
    Dict,
    TypeVar,
    Union,
    TYPE_CHECKING,
    Iterable,
    Iterator,
    Set,
    Optional,
    Mapping,
    _GenericAlias,  # type: ignore
)

from typic import gen, util, checks
from typic.compat import ForwardRef, evaluate_forwardref, Protocol
from .error import ConstraintValueError

if TYPE_CHECKING:  # pragma: nocover
    from typic.constraints.factory import ConstraintsT  # noqa: F401


__all__ = (
    "BaseConstraints",
    "EnumConstraints",
    "LiteralConstraints",
    "MultiConstraints",
    "TypeConstraints",
    "ValidatorT",
    "ValidateT",
    "ConstraintsProtocolT",
    "VT",
)

VT = TypeVar("VT")
"""A generic type-var for values passed to a validator."""
AssertionsT = List[str]
ContextT = Dict[str, Any]
_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)


class ValidatorT(Protocol[_T_co]):
    """The expected signature of a value validator."""

    def __call__(
        self, value: Any, *, field: str = None, **kwargs
    ) -> Tuple[bool, _T_co]:
        ...


class ValidateT(Protocol[_T_co]):
    """The signature of the public validate callable for a Constraint."""

    def __call__(self, value: Any, *, field: str = None) -> _T_co:
        ...


class ConstraintsProtocolT(Protocol[_T]):
    """The guaranteed protocol for a Constraints instance."""

    type: Type
    nullable: bool
    coerce: bool
    name: Optional[str]
    validator: ValidatorT[_T]
    validate: ValidateT[_T]

    def for_schema(self, *, with_type: bool = False) -> Dict[str, Any]:
        ...


class _AbstractConstraints(abc.ABC):
    type: ClassVar[Type[Any]]
    VALUE = "value"
    VALTNAME = "valtname"
    FIELD = "field"
    FNAME = "fieldname"
    NULLABLES = (None, Ellipsis)

    __slots__ = ("__dict__",)
    __ignore_repr__ = frozenset(("type",))

    def __init_subclass__(cls, **kwargs):
        cls.__ignore_repr__ = _AbstractConstraints.__ignore_repr__ | cls.__ignore_repr__

    @util.cached_property
    @reprlib.recursive_repr()
    def __str(self) -> str:
        fields = [f"type={self.type_qualname}"]
        for f in dataclasses.fields(self):
            if f.name in self.__ignore_repr__:
                continue

            val = getattr(self, f.name)
            if (val or val in {False, 0}) and f.repr:
                fields.append(f"{f.name}={val!r}")
        return f"({', '.join(fields)})"

    def __str__(self) -> str:
        return self.__str

    def __repr__(self):
        return self.__str

    def define(self, block: gen.Block, name: str) -> gen.Function:
        f: gen.Function = block.f(
            name,
            block.param(self.VALUE, annotation="VT"),
            block.param(
                "field",
                annotation=str,
                kind=gen.ParameterKind.KEYWORD_ONLY,  # type: ignore
                default=None,
            ),
        )
        return f

    @util.cached_property
    def type_name(self) -> str:
        return util.get_name(self.type)

    @util.cached_property
    def type_qualname(self) -> str:
        return util.get_qualname(self.type)

    def _get_validator_name(self) -> str:
        return util.get_defname("validator", self)

    @util.cached_property
    @abc.abstractmethod
    def validator(self) -> ValidatorT:  # pragma: nocover
        ...

    def validate(self, value: VT, *, field: str = None) -> VT:
        """Validate that an incoming value meets the given constraints.

        Notes
        -----
        Some constraints may mutate the incoming value to conform, so the value is always
        returned.

        Raises
        ------
        :py:class:`ConstraintValueError`
            An error inheriting from :py:class:`ValueError` indicating the input is not
            valid.
        :py:class:`ConstraintSyntaxError`
            An error inheriting from :py:class:`SyntaxError` indicating the constraint
            configuration is invalid.
        """
        try:
            valid, value = self.validator(value, field=field)
        except AttributeError:
            valid, value = False, value

        if not valid:
            field = f"{field}:" if field is not None else "Given"
            raise ConstraintValueError(
                f"{field} value <{value!r}> fails constraints: {self}"
            ) from None
        return value

    @abc.abstractmethod
    def for_schema(
        self, *, with_type: bool = False
    ) -> Dict[str, Any]:  # pragma: nocover
        ...


class InstanceCheck(enum.IntEnum):
    """Flags for instance check methods.

    See Also
    --------
    :py:class:`~typic.constraints.mapping.MappingConstraints`
    """

    IS = 0
    """Allows for short-circuiting validation if `isinstance(value, <type>) is True`.

    Otherwise, perform additional checks to see if we can treat this as valid.
    """
    NOT = 1
    """Allows for short-circuiting validation if `isinstance(value, <type>) is False`.

    Otherwise, we must perform additional checks.
    """


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)  # type: ignore
class BaseConstraints(_AbstractConstraints):
    """A base constraints object. Shouldn't be used directly.

    Notes
    -----
    Inheritors of :py:class:`Constraints` should conform to JSON Schema type-constraints.

    See Also
    --------
    :py:class:`~typic.types.constraints.TextConstraints`
    :py:class:`~typic.types.constraints.NumberConstraints`
    :py:class:`~typic.types.constraints.ArrayConstraints`
    :py:class:`~typic.types.constraints.MappingConstraints`
    """

    type: ClassVar[Type[Any]]
    """The target type for this constraint."""
    instancecheck: ClassVar[InstanceCheck] = InstanceCheck.NOT
    """The method to use when short-circuting with instance checks."""
    nullable: bool = False
    """Whether to allow null values."""
    coerce: bool = False
    """Whether additional coercion should be done after validation.

    Even in `strict` mode, we may still want to coerce after validation.
    """
    name: Optional[str] = None

    __ignore_repr__ = frozenset(("coerce", "name", "instancecheck"))

    def __post_init__(self):
        self.validator

    def _build_validator(
        self, func: gen.Block, context: ContextT, assertions: AssertionsT
    ) -> ContextT:
        if assertions:
            self._build_assertions(func=func, assertions=assertions)
        return context

    def _build_assertions(self, func: gen.Block, assertions: AssertionsT):
        check = " and ".join(assertions)
        with func.b(f"if not ({check}):") as b:
            b.l(f"return False, {self.VALUE}")

    def _check_syntax(self):
        return

    def _get_assertions(self) -> AssertionsT:
        raise NotImplementedError

    def _compile_validator(self) -> ValidatorT:
        func_name = self._get_validator_name()
        origin = util.origin(self.type)
        type_name = self.type_name
        self._check_syntax()
        assertions = self._get_assertions()
        context: ContextT = {type_name: self.type}
        with gen.Block() as main:
            with self.define(main, func_name) as f:
                # This is a signal that -*-anything can happen...-*-
                if origin in {Any, Signature.empty}:
                    f.l(f"return True, {self.VALUE}")
                    return main.compile(name=func_name)
                f.l(f"{self.VALTNAME} = {type_name!r}")
                f.l(f"{self.FNAME} = {self.VALTNAME} if field is None else field")
                # Short-circuit validation if the value isn't the correct type.
                if self.instancecheck == InstanceCheck.IS:
                    line = f"if isinstance({self.VALUE}, {type_name}):"
                    if self.nullable:
                        line = (
                            f"if {self.VALUE} in {self.NULLABLES} "
                            f"or isinstance({self.VALUE}, {type_name}):"
                        )
                    with f.b(line, **context) as b:  # type: ignore
                        b.l(f"return True, {self.VALUE}")
                else:
                    if self.nullable:
                        with f.b(f"if {self.VALUE} in {self.NULLABLES}:") as b:
                            b.l(f"return True, {self.VALUE}")
                    line = f"if not isinstance({self.VALUE}, {type_name}):"
                    with f.b(line, **context) as b:  # type: ignore
                        b.l(f"return False, {self.VALUE}")
                context = self._build_validator(
                    f, context=context, assertions=assertions
                )
                f.namespace.update(context)
                f.localize_context(*context)
                f.l(f"return True, {self.VALUE}")
        return main.compile(name=func_name)

    @util.cached_property
    def validator(self) -> ValidatorT:
        """Accessor for the generated validator.

        Validators are generated code based upon the constraint syntax provided.
        The resulting code is about as optimized as you can get in Python, since
        the constraint values are localized to the validator and only the checks
        that the validator performs are the checks that were set. There is no
        computation time wasted on deciding whether or not to perform a validation.
        """
        validator = self._compile_validator()
        return validator


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class MultiConstraints(_AbstractConstraints):
    """A container for multiple constraints for a single field."""

    constraints: Tuple[ConstraintsProtocolT, ...]
    coerce: bool = False
    name: Optional[str] = None
    tag: Optional[util.TaggedUnion] = None

    @util.cached_property
    def nullable(self) -> bool:
        return any(x.nullable for x in self.constraints)

    @util.cached_property
    def __str(self) -> str:  # type: ignore
        constraints = f"({', '.join((str(c) for c in self.constraints))})"
        return f"(constraints={constraints}, nullable={self.nullable})"

    def __str__(self) -> str:
        return self.__str

    def __repr__(self):
        return self.__str

    def __type(self) -> Iterator[Type]:
        to_flatten: Set[Union[Type, Iterable[Type]]]
        to_flatten = {x.type for x in self.constraints}
        while to_flatten:
            t: Union[Type, Iterable[Type]] = to_flatten.pop()
            if isinstance(t, Iterable):
                to_flatten.update(t)
                continue
            yield t

    @util.cached_property
    def type(self) -> Tuple[Type, ...]:  # type: ignore
        return (*self.__type(),)

    @util.cached_property
    def type_name(self) -> str:
        return f"({', '.join(util.get_name(t) for t in self.type)})"

    @util.cached_property
    def type_qualname(self) -> str:
        return f"({', '.join(util.get_qualname(t) for t in self.type)})"

    @util.cached_property
    def validator(self) -> ValidatorT:
        """Accessor for the generated multi-validator.

        Validators are keyed by the origin-type of :py:class:`BaseConstraints` inheritors.

        If a value does not match any origin-type, as reported by :py:func:`typic.origin`,
        then we will report the value as invalid.
        """
        func_name = self._get_validator_name()
        vmap = util.TypeMap({c.type: c for c in self.constraints})
        ns = {"tag": self.tag and self.tag.tag, "empty": util.empty}
        with gen.Block(ns) as main:
            with self.define(main, func_name) as f:
                if not vmap:
                    f.l(f"return True, {self.VALUE}")
                else:
                    f.l(f"{self.VALTNAME} = {self.VALUE}.__class__")
                    if self.nullable:
                        with f.b(f"if {self.VALUE} is None:") as b:
                            b.l(f"return True, {self.VALUE}")
                    if self.tag:
                        validators = {
                            value: vmap[t] for value, t in self.tag.types_by_values
                        }
                        f.namespace.update(vmap=validators)
                        with f.b(
                            f"if issubclass({self.VALTNAME}, Mapping):",
                            Mapping=collections.abc.Mapping,
                        ) as b:
                            b.l(f"tag_value = {self.VALUE}.get(tag, empty)")
                        with f.b("else:") as b:
                            b.l(f"tag_value = getattr({self.VALUE}, tag, empty)")
                        f.l(
                            f"valid, {self.VALUE} = "
                            f"(True, vmap[tag_value].validate({self.VALUE}, field=field)) "
                            f"if tag_value in vmap else (False, {self.VALUE})"
                        )
                    else:
                        vmap = util.TypeMap(
                            {util.origin(t): v for t, v in vmap.items()}
                        )
                        f.namespace.update(vmap=vmap)
                        f.l(f"v = vmap.get_by_parent({self.VALTNAME}, None)")
                        f.l(
                            f"valid, {self.VALUE} = (True, v.validate(value, field=field)) "
                            f"if v else (False, value)"
                        )
                    f.l(f"return valid, {self.VALUE}")
        validator = main.compile(name=func_name)
        return validator  # type: ignore

    def for_schema(self, *, with_type: bool = False) -> dict:
        scheme: dict = {
            "anyOf": [x.for_schema(with_type=True) for x in self.constraints],
        }
        if self.name:
            scheme["name"] = self.name
        return scheme


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class TypeConstraints(_AbstractConstraints):
    """A container for simple types. Validation is limited to instance checks.

    Examples
    --------
    >>> import typic
    >>> tc = typic.get_constraints(typic.AbsoluteURL)
    >>> tc.validate("http://foo.bar")
    Traceback (most recent call last):
        ...
    typic.constraints.error.ConstraintValueError: Given value <'http://foo.bar'> fails constraints: (type=AbsoluteURL, nullable=False)
    >>> tc = typic.get_constraints(typic.AbsoluteURL, nullable=True)
    >>> tc.validate(None)

    >>>
    """

    type: Type[Any]  # type: ignore
    """The type to check for."""
    nullable: bool = False
    """Whether this constraint can allow null values."""
    name: Optional[str] = None

    def __post_init__(self):
        self.validator

    @util.cached_property
    def validator(self) -> ValidatorT:
        ns = dict(__t=self.type, VT=VT)
        func_name = self._get_validator_name()
        with gen.Block(ns) as main:
            with self.define(main, func_name) as f:
                # Standard instancecheck is default.
                check = "isinstance(value, __t)"
                retval = "value"
                # Have to allow nulls if its nullable.
                if self.nullable:
                    check = "(value is None or isinstance(value, __t))"
                elif self.type is Any:
                    check = "True"
                f.l(f"{gen.Keyword.RET} {check}, {retval}")

        validator: ValidatorT = main.compile(name=func_name, ns=ns)
        return validator

    def for_schema(self, *, with_type: bool = False) -> Dict[str, Any]:
        if self.name:
            return {"title": self.name}
        return {}


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class EnumConstraints(_AbstractConstraints):
    type: Type[enum.Enum]  # type: ignore
    """The enum to check for."""
    nullable: bool = False
    """Whether this constraint can allow null values."""
    coerce: bool = True
    name: Optional[str] = None

    def __post_init__(self):
        self.validator

    @util.cached_property
    def __str(self) -> str:
        values = (*(x.value for x in self.type),)
        return (
            f"(type={self.name or self.type.__name__}, "
            f"values={values}, "
            f"nullable={self.nullable})"
        )

    def __str__(self) -> str:
        return self.__str

    def __repr__(self):
        return self.__str

    @util.cached_property
    def validator(self) -> ValidatorT:
        ns = dict(__t=self.type, VT=VT, __values=(*self.type,))
        func_name = self._get_validator_name()
        with gen.Block(ns) as main:
            with self.define(main, func_name) as f:
                if self.nullable:
                    with f.b(f"if value in {self.NULLABLES}:") as b:
                        b.l(f"{gen.Keyword.RET} True, value")
                # This is O(N), but so is casting to the enum
                # And handling a ValueError is an order of magnitude heavier
                f.l(f"{gen.Keyword.RET} value in __values, value")

        validator: ValidatorT = main.compile(name=func_name, ns=ns)
        return validator

    def for_schema(self, *, with_type: bool = False) -> Dict[str, Any]:
        if self.name:
            return {"title": self.name}
        return {}


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class LiteralConstraints(_AbstractConstraints):
    type: _GenericAlias  # type: ignore  # real-type: Literal
    nullable: bool = False
    coerce: bool = False
    name: Optional[str] = None

    def __post_init__(self):
        self.validator

    @property
    def values(self):
        return util.get_args(self.type)

    @util.cached_property
    def __str(self) -> str:
        return f"(type=Literal, " f"values={self.values}, " f"nullable={self.nullable})"

    def __str__(self) -> str:
        return self.__str

    def __repr__(self):
        return self.__str

    @util.cached_property
    def validator(self) -> ValidatorT:
        ns = dict(__t=self.type, VT=VT, __values=frozenset(self.values))
        func_name = self._get_validator_name()
        with gen.Block(ns) as main:
            with self.define(main, func_name) as f:
                if self.nullable:
                    with f.b(f"if value in {self.NULLABLES}:") as b:
                        b.l(f"{gen.Keyword.RET} True, value")
                with f.b("try:") as b:
                    b.l(f"{gen.Keyword.RET} value in __values, value")
                with f.b("except TypeError:") as b:
                    b.l(f"{gen.Keyword.RET} False, value")

        validator: ValidatorT = main.compile(name=func_name, ns=ns)
        return validator

    def for_schema(self, *, with_type: bool = False) -> Dict[str, Any]:
        if self.name:
            return {"title": self.name}
        return {}


class DelayedConstraints:
    __slots__ = "t", "nullable", "name", "factory", "_constraints"

    def __init__(self, t: Type, nullable: bool, name: Optional[str], factory: Callable):
        self.t = t
        self.nullable = nullable
        self.name = name
        self.factory = factory
        self._constraints: Optional["_AbstractConstraints"] = None

    def _evaluate_contraints(self):
        type = self.t
        if checks.isoptionaltype(type):
            args = util.get_args(type)[:-1]
            type = args[0] if len(args) == 1 else Union[args]
            self.nullable = True
            self.t = type
        c = self.factory(type, nullable=self.nullable, name=self.name)
        return c

    @property
    def constraints(self) -> "_AbstractConstraints":
        if self._constraints is None:
            self._constraints = self._evaluate_contraints()
        return self._constraints

    def validate(self, value: VT, *, field: str = None):
        return self.constraints.validate(value, field=field)

    def __getattr__(self, item):
        return self.constraints.__getattribute__(item)


class ForwardDelayedConstraints:
    __slots__ = (
        "ref",
        "module",
        "localns",
        "nullable",
        "name",
        "factory",
        "_constraints",
    )

    def __init__(
        self,
        ref: ForwardRef,
        module: str,
        localns: Mapping,
        nullable: bool,
        name: Optional[str],
        factory: Callable,
    ):
        self.ref = ref
        self.module = module
        self.localns = localns
        self.nullable = nullable
        self.name = name
        self.factory = factory
        self._constraints: Optional["_AbstractConstraints"] = None

    def _evaluate_contraints(self):
        globalns = sys.modules[self.module].__dict__.copy()
        try:
            type = evaluate_forwardref(self.ref, globalns or {}, self.localns or {})
        except NameError as e:  # pragma: nocover
            warnings.warn(
                f"Counldn't resolve forward reference: {e}. "
                f"Make sure this type is available in {self.module}."
            )
            type = object  # make it a no-op
        if checks.isoptionaltype(type):
            args = util.get_args(type)[:-1]
            type = args[0] if len(args) == 1 else Union[args]
            self.nullable = True
        c = self.factory(type, nullable=self.nullable, name=self.name)
        return c

    @property
    def constraints(self) -> "_AbstractConstraints":
        if self._constraints is None:
            self._constraints = self._evaluate_contraints()
        return self._constraints

    def validate(self, value: VT, *, field: str = None):
        return self.constraints.validate(value, field=field)

    def __getattr__(self, item):
        return self.constraints.__getattribute__(item)



from __future__ import annotations

from typing import Optional, Callable, List, Union
from typing_extensions import Final

from mypy.semanal import SemanticAnalyzer
from mypy.semanal_shared import set_callable_name
from mypy.typeops import TypingType

from mypy.nodes import (
    MDEF,
    Argument,
    SymbolTableNode,
    Var,
    TypeVarExpr,
    Decorator,
    NameExpr,
    PassStmt,
    Block,
    FuncDef,
    FuncBase,
    SymbolNode,
    ARG_POS,
    ARG_NAMED_OPT,
)
from mypy.plugin import ClassDefContext, Plugin, MethodContext
from mypy.plugins.common import _get_decorator_bool_argument
from mypy.types import (
    Type,
    TypeVarDef,
    TypeVarType,
    AnyType,
    TypeOfAny,
    CallableType,
    TypeType,
)
from mypy.plugins.dataclasses import DataclassTransformer
from mypy.typevars import fill_typevars
from mypy.util import get_unique_redefinition_name

typic_class_maker_decorators = {
    "klass",
    "settings",
    "typic.klass",
    "typic.klass.klass",
    "typic.settings",
    "typic.api.settings",
}  # type: Final

typic_class_maker = "typic.api.wrap_cls"

typic_methods = {
    "primitive",
    "transmute",
    "validate",
    "schema",
    "tojson",
}  # type: Final

SELF_TVAR_NAME = "_TT"


def typic_method_callback(ctx: MethodContext) -> Type:
    return ctx.default_return_type


def plugin(version: str) -> "TypingType[Plugin]":
    """
    `version` is the mypy version string
    We might want to use this to print a warning if the mypy version being used is
    newer, or especially older, than we expect (or need).
    """
    return TypicPlugin


class TypicPlugin(Plugin):
    def get_method_hook(
        self, fullname: str
    ) -> Optional[Callable[[MethodContext], Type]]:
        if any(fullname.endswith(x) for x in typic_methods):
            return typic_method_callback
        return None

    def get_class_decorator_hook(
        self, fullname: str
    ) -> Optional[Callable[[ClassDefContext], None]]:
        if fullname in typic_class_maker_decorators:
            return typic_klass_maker_callback
        return None


class TypicTransformer:
    """A transformer for allowing MyPy to recognize typical klasses."""

    def __init__(self, ctx: ClassDefContext):
        self._ctx = ctx
        self.dclass = DataclassTransformer(self._ctx)

    def transform(self) -> None:
        self.dclass.transform()
        self._add_self_tvar_expr(self._ctx)
        jsonschema = _get_decorator_bool_argument(self._ctx, "jsonschema", True)
        if jsonschema:
            self.add_schema_method()
        self.add_primitive_method()
        self.add_transmute_method()
        self.add_validate_method()
        self.add_json_method()
        self.add_translate_method()
        self.add_iterate_method()
        self.add_iter_method()

    @staticmethod
    def _get_tvar_name(name: str, info) -> str:
        return f"{info.fullname}.{name}"

    def _get_tvar_self_name(self) -> str:
        return self._get_tvar_name(SELF_TVAR_NAME, self._ctx.cls.info)

    def _add_tvar_expr(self, name: str, ctx):
        info = ctx.cls.info
        obj_type = ctx.api.named_type("__builtins__.object")
        self_tvar_expr = TypeVarExpr(
            name, self._get_tvar_name(name, info), [], obj_type
        )
        info.names[name] = SymbolTableNode(MDEF, self_tvar_expr)

    def _add_self_tvar_expr(self, ctx):
        self._add_tvar_expr(SELF_TVAR_NAME, ctx)

    def _get_tvar_def(self, name: str, ctx):
        obj_type = ctx.api.named_type("__builtins__.object")
        return TypeVarDef(
            SELF_TVAR_NAME, self._get_tvar_name(name, ctx.cls.info), -1, [], obj_type
        )

    def add_schema_method(self):
        ctx = self._ctx
        api: SemanticAnalyzer = ctx.api
        return_type_info: SymbolTableNode = api.lookup_fully_qualified(
            "typic.SchemaReturnT"
        )
        self_tvar_def = self._get_tvar_def(SELF_TVAR_NAME, ctx)
        arg_type = api.named_type("__builtins__.bool")
        arg = Argument(Var("primitive", arg_type), arg_type, None, ARG_NAMED_OPT)
        add_method(
            ctx,
            "schema",
            args=[arg],
            return_type=return_type_info.node.target,
            self_type=TypeVarType(self_tvar_def),
            tvar_def=self_tvar_def,
            is_classmethod=True,
        )

    def add_primitive_method(self):
        ctx = self._ctx
        self_tvar_def = self._get_tvar_def(SELF_TVAR_NAME, ctx)
        bool_type = ctx.api.named_type("__builtins__.bool")
        arg = Argument(Var("lazy", bool_type), bool_type, None, ARG_NAMED_OPT)
        add_method(
            ctx,
            "primitive",
            args=[arg],
            return_type=AnyType(TypeOfAny.unannotated),
            self_type=TypeVarType(self_tvar_def),
            tvar_def=self_tvar_def,
        )

    def add_json_method(self):
        ctx = self._ctx
        self_tvar_def = self._get_tvar_def(SELF_TVAR_NAME, ctx)
        bool_type = ctx.api.named_type("__builtins__.bool")
        int_type = ctx.api.named_type("__builtins__.int")
        str_type = ctx.api.named_type("__builtins__.str")
        indent = Argument(Var("indent", int_type), int_type, None, ARG_NAMED_OPT)
        ensure = Argument(
            Var("ensure_ascii", bool_type), bool_type, None, ARG_NAMED_OPT
        )
        add_method(
            ctx,
            "tojson",
            args=[indent, ensure],
            return_type=str_type,
            self_type=TypeVarType(self_tvar_def),
            tvar_def=self_tvar_def,
        )

    def add_validate_method(self):
        ctx = self._ctx
        self_tvar_def = self._get_tvar_def(SELF_TVAR_NAME, ctx)
        arg_type = AnyType(TypeOfAny.explicit)
        arg = Argument(Var("obj", arg_type), arg_type, None, ARG_POS)
        add_method(
            ctx,
            "validate",
            args=[arg],
            return_type=TypeVarType(self_tvar_def),
            tvar_def=self_tvar_def,
            is_staticmethod=True,
        )

    def add_transmute_method(self):
        ctx = self._ctx
        self_tvar_def = self._get_tvar_def(SELF_TVAR_NAME, ctx)
        arg_type = AnyType(TypeOfAny.explicit)
        arg = Argument(Var("obj", arg_type), arg_type, None, ARG_POS)
        add_method(
            ctx,
            "transmute",
            args=[arg],
            return_type=TypeVarType(self_tvar_def),
            tvar_def=self_tvar_def,
            is_staticmethod=True,
        )

    def add_translate_method(self):
        ctx = self._ctx
        self_tvar_def = self._get_tvar_def(SELF_TVAR_NAME, ctx)
        r_type = AnyType(TypeOfAny.explicit)
        arg_type = TypeType(r_type)
        arg = Argument(Var("target", arg_type), arg_type, None, ARG_POS)
        add_method(
            ctx,
            "translate",
            args=[arg],
            return_type=r_type,
            self_type=TypeVarType(self_tvar_def),
            tvar_def=self_tvar_def,
        )

    def add_iterate_method(self):
        ctx = self._ctx
        self_tvar_def = self._get_tvar_def(SELF_TVAR_NAME, ctx)
        r_type = AnyType(TypeOfAny.explicit)
        bool_type = ctx.api.named_type("__builtins__.bool")
        arg = Argument(Var("values", bool_type), bool_type, None, ARG_NAMED_OPT)
        add_method(
            ctx,
            "iterate",
            args=[arg],
            return_type=r_type,
            self_type=TypeVarType(self_tvar_def),
            tvar_def=self_tvar_def,
        )

    def add_iter_method(self):
        ctx = self._ctx
        self_tvar_def = self._get_tvar_def(SELF_TVAR_NAME, ctx)
        r_type = AnyType(TypeOfAny.explicit)
        bool_type = ctx.api.named_type("__builtins__.bool")
        arg = Argument(Var("values", bool_type), bool_type, None, ARG_NAMED_OPT)
        add_method(
            ctx,
            "__iter__",
            args=[arg],
            return_type=r_type,
            self_type=TypeVarType(self_tvar_def),
            tvar_def=self_tvar_def,
        )


def typic_klass_maker_callback(ctx: ClassDefContext) -> None:
    transformer = TypicTransformer(ctx)
    transformer.transform()


def add_method(
    ctx: ClassDefContext,
    name: str,
    args: List[Argument],
    return_type: Type,
    self_type: Optional[Type] = None,
    tvar_def: Optional[TypeVarDef] = None,
    is_classmethod: bool = False,
    is_new: bool = False,
    is_staticmethod: bool = False,
) -> None:
    """
    Adds a new method to a class.
    This can be dropped if/when https://github.com/python/mypy/issues/7301 is merged
    """
    info = ctx.cls.info

    # First remove any previously generated methods with the same name
    # to avoid clashes and problems in the semantic analyzer.
    if name in info.names:
        sym = info.names[name]
        if sym.plugin_generated and isinstance(sym.node, FuncDef):
            ctx.cls.defs.body.remove(sym.node)

    self_type = self_type or fill_typevars(info)
    if is_classmethod or is_new:
        first = [
            Argument(Var("_cls"), TypeType.make_normalized(self_type), None, ARG_POS)
        ]
    elif is_staticmethod:
        first = []
    else:
        self_type = self_type or fill_typevars(info)
        first = [Argument(Var("self"), self_type, None, ARG_POS)]
    args = first + args
    arg_types, arg_names, arg_kinds = [], [], []
    for arg in args:
        assert arg.type_annotation, "All arguments must be fully typed."
        arg_types.append(arg.type_annotation)
        arg_names.append(get_name(arg.variable))
        arg_kinds.append(arg.kind)

    function_type = ctx.api.named_type("__builtins__.function")
    signature = CallableType(
        arg_types, arg_kinds, arg_names, return_type, function_type
    )
    if tvar_def:
        signature.variables = [tvar_def]

    func = FuncDef(name, args, Block([PassStmt()]))
    func.info = info
    func.type = set_callable_name(signature, func)
    func.is_class = is_classmethod
    func.is_static = is_staticmethod
    func._fullname = get_fullname(info) + "." + name
    func.line = info.line

    # NOTE: we would like the plugin generated node to dominate, but we still
    # need to keep any existing definitions so they get semantically analyzed.
    if name in info.names:
        # Get a nice unique name instead.
        r_name = get_unique_redefinition_name(name, info.names)
        info.names[r_name] = info.names[name]

    if is_classmethod or is_staticmethod:
        func.is_decorated = True
        v = Var(name, func.type)
        v.info = info
        v._fullname = func._fullname
        if is_classmethod:
            v.is_classmethod = True
            dec = Decorator(func, [NameExpr("classmethod")], v)
        else:
            v.is_staticmethod = True
            dec = Decorator(func, [NameExpr("staticmethod")], v)

        dec.line = info.line
        sym = SymbolTableNode(MDEF, dec)
    else:
        sym = SymbolTableNode(MDEF, func)
    sym.plugin_generated = True

    info.names[name] = sym
    info.defn.defs.body.append(func)


def get_fullname(x: Union[FuncBase, SymbolNode]) -> str:
    """
    Used for compatibility with mypy 0.740; can be dropped once support for 0.740 is dropped.
    """
    fn = x.fullname
    if callable(fn):  # pragma: nocover
        return fn()
    return fn


def get_name(x: Union[FuncBase, SymbolNode]) -> str:
    """
    Used for compatibility with mypy 0.740; can be dropped once support for 0.740 is dropped.
    """
    fn = x.name
    if callable(fn):  # pragma: nocover
        return fn()
    return fn



from __future__ import annotations

import inspect
from operator import methodcaller
from typing import (
    TYPE_CHECKING,
    Type,
    Mapping,
    Tuple,
    Optional,
    Set,
    Dict,
    Any,
    Callable,
    Iterator,
    Union,
    Iterable,
)

from typic.checks import (
    ismappingtype,
    isiterabletype,
    isliteral,
    isnamedtuple,
    isiteratortype,
)
from typic.compat import lru_cache
from typic.gen import Block, Keyword, ParameterKind
from typic.util import (
    cached_type_hints,
    cached_simple_attributes,
    safe_get_params,
    get_unique_name,
    get_defname,
    get_name,
)

if TYPE_CHECKING:
    from .common import Annotation, TranslatorT, SerdeProtocol
    from .resolver import Resolver


_itemscaller = methodcaller("items")
_valuescaller = methodcaller("values")
_iter = iter


class TranslatorTypeError(TypeError):
    ...


class TranslatorValueError(ValueError):
    ...


class TranslatorFactory:
    """Translation protocol factory for higher-order objects.

    Notes
    -----
    For lower-order objects this will be functionally the same as a serializer.
    """

    KWD_KINDS = frozenset(
        {inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.VAR_KEYWORD}
    )
    POS_KINDS = frozenset(
        {inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.VAR_POSITIONAL}
    )
    VAR_KINDS = frozenset(
        {inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD}
    )

    def __init__(self, resolver: Resolver):
        self.resolver = resolver

    def sig_is_undef(self, params: Mapping[str, inspect.Parameter]) -> bool:
        return (not params) or {x.kind for x in params.values()}.issubset(
            self.VAR_KINDS
        )

    def kw_only(self, params: Mapping[str, inspect.Parameter]) -> bool:
        return not any(x.kind in self.POS_KINDS for x in params.values())

    def pos_only(self, params: Mapping[str, inspect.Parameter]) -> bool:
        return not any(x.kind in self.KWD_KINDS for x in params.values())

    @staticmethod
    def required_fields(params: Mapping[str, inspect.Parameter]) -> Set[str]:
        return {x for x, y in params.items() if y.default is y.empty}

    @staticmethod
    def _fields_from_hints(
        kind: ParameterKind, hints: Mapping[str, Type], exclude: Set[str]
    ) -> Mapping[str, inspect.Parameter]:
        return {
            x: inspect.Parameter(x, kind, annotation=y)
            for x, y in hints.items()
            if x not in exclude
        }

    @staticmethod
    def _fields_from_attrs(
        kind: ParameterKind, attrs: Tuple[str, ...], exclude: Set[str]
    ):
        return {x: inspect.Parameter(x, kind) for x in attrs if x not in exclude}

    @lru_cache(maxsize=None)
    def get_fields(
        self, type: Type, as_source: bool = False, exclude: Iterable[str] = ()
    ) -> Optional[Mapping[str, inspect.Parameter]]:
        """Get the fields for the given type.

        Notes
        -----
        We want this to be the type's signature, we really do. But if for some reason we
        can't make that happen, we fallback to a few known, semi-reliable methods for
        making this happen.
        """
        # Try first with the signature of the target if this is the target type
        exclude = {*exclude}
        params = safe_get_params(type)
        undefined = self.sig_is_undef(params)
        if not as_source and not undefined:
            return params
        # Now we start building a fake signature
        k: ParameterKind = inspect.Parameter.POSITIONAL_OR_KEYWORD
        # **kwargs
        if self.kw_only(params):
            k = inspect.Parameter.KEYWORD_ONLY
        # *args
        elif self.pos_only(params):
            k = inspect.Parameter.POSITIONAL_ONLY
        # Fetch any type hints and try to use those.
        hints = cached_type_hints(type)
        if hints:
            return self._fields_from_hints(k, hints, exclude)
        # Fallback to the target object's defined attributes
        # This will basically work for ORM models, Pydantic models...
        # Anything that defines the instance using the class body.
        attrs = cached_simple_attributes(type)
        if attrs:
            return self._fields_from_attrs(k, attrs, exclude)
        # Can't be done.
        return None if undefined else {f: params[f] for f in params.keys() - exclude}

    @lru_cache(maxsize=None)
    def iterator(
        self,
        type: Type,
        values: bool = False,
        relaxed: bool = False,
        exclude: Tuple[str, ...] = (),
    ) -> IteratorT:
        """Get an iterator function for a given type, if possible."""

        if ismappingtype(type):
            iter = _valuescaller if values else _itemscaller
            return iter

        if isiterabletype(type) and not isnamedtuple(type):
            return _iter

        fields = self.get_fields(type, as_source=True, exclude=exclude) or {}

        if not fields and not relaxed:
            raise TranslatorTypeError(
                f"Cannot get iterator for type {type!r}, unable to determine fields."
            ) from None

        func_name = get_defname("iterator", (type, values))
        oname = "o"
        ctx: dict = {}
        with Block(ctx) as main:
            with main.f(func_name, Block.p(oname)) as func:
                if fields:
                    if values:
                        for f in fields:
                            func.l(f"{Keyword.YLD} {oname}.{f}")
                    else:
                        for f in fields:
                            func.l(f"{Keyword.YLD} {f!r}, {oname}.{f}")
                else:
                    func.l(f"{Keyword.YLD}")

        return main.compile(name=func_name, ns=ctx)

    @staticmethod
    def _get_name(source: Type, target: Type) -> str:
        return get_defname("translator", (source, target))

    @staticmethod
    def _iter_field_assigns(
        fields: Mapping[str, inspect.Parameter],
        oname: str,
        protos: Mapping[str, SerdeProtocol],
        ctx: Dict[str, Any],
    ):
        for f, p in fields.items():
            fset = f"{oname}.{f}"
            if f in protos:
                deser_name = f"{f}_deser"
                proto = protos[f]
                ctx[deser_name] = proto.transmute
                fset = f"{deser_name}({fset})"
            if p.kind != p.POSITIONAL_ONLY:
                fset = f"{f}={fset}"
            yield fset

    def _compile_iterable_translator(self, source: Type, target: Type) -> TranslatorT:
        func_name = self._get_name(source, target)
        target_name = get_name(target)
        oname = "o"
        ismapping = ismappingtype(target)
        iterator = self.iterator(source, not ismapping)
        ctx = {"iterator": iterator, target_name: target}
        with Block(ctx) as main:
            with main.f(func_name, Block.p(oname)) as func:
                retval = f"iterator({oname})"
                if not isiteratortype(target):
                    retval = f"{target_name}({retval})"
                func.l(f"{Keyword.RET} {retval}")
        return main.compile(name=func_name)

    @lru_cache(maxsize=None)
    def _compile_translator(
        self, source: Type, target: Type, exclude: Tuple[str, ...] = ()
    ) -> TranslatorT:
        if isliteral(target):
            raise TranslatorTypeError(
                f"Cannot translate to literal type: {target!r}. "
            ) from None
        if isliteral(source):
            raise TranslatorTypeError(
                f"Cannot translate from literal type: {source!r}. "
            ) from None
        # Get the target fields for translation.
        target_fields = self.get_fields(target)
        if target_fields is None:
            if isiterabletype(target):
                return self._compile_iterable_translator(source, target)
            raise TranslatorTypeError(
                f"Cannot translate to type {target!r}. "
                f"Unable to determine target fields."
            ) from None

        # Ensure that the target fields are a subset of the source fields.
        # We treat the target fields as the parameters for the target,
        # so this must be true.
        fields = self.get_fields(source, as_source=True, exclude=exclude) or {}
        fields_to_pass = {x: fields[x] for x in fields.keys() & target_fields.keys()}
        required = self.required_fields(target_fields)
        if not required.issubset(fields_to_pass.keys()):
            diff = (*(required - fields.keys()),)
            raise TranslatorValueError(
                f"{source!r} can't be translated to {target!r}. "
                f"Source is missing required fields: {diff}."
            ) from None
        protocols = self.resolver.protocols(target)

        # Build the translator.
        anno_name = get_unique_name(source)
        target_name = get_unique_name(target)
        func_name = self._get_name(source, target)
        oname = "o"
        ctx: Dict[str, Any] = {target_name: target, anno_name: source}
        with Block(ctx) as main:
            with main.f(func_name, Block.p(oname)) as func:
                args = ", ".join(
                    self._iter_field_assigns(fields_to_pass, oname, protocols, ctx)
                )
                func.l(f"{Keyword.RET} {target_name}({args})")
        trans = main.compile(name=func_name, ns=ctx)
        return trans

    def factory(
        self, annotation: "Annotation", target: Type, exclude: Tuple[str, ...] = ()
    ) -> TranslatorT:
        """Generate a translator for :py:class:`typic.Annotation` -> ``type``."""
        exclude = (*(exclude or annotation.serde.flags.exclude),)
        return self._compile_translator(annotation.resolved, target, exclude=exclude)


IteratorT = Union[Callable[[Any], Iterator[Any]], Callable[[Any], Tuple[str, Any]]]



from __future__ import annotations

import builtins
import inspect
import os
from typing import TypeVar, Type, Any, TYPE_CHECKING, Mapping

from typic import types
from typic.checks import STDLIB_TYPES
from typic.serde import common
from typic.util import get_name


if TYPE_CHECKING:  # pragma: nocover
    from typic.serde.resolver import Resolver


_ET = TypeVar("_ET")


class EnvironmentValueError(ValueError):
    ...


class EnvironmentTypeError(TypeError):
    ...


class Environ:
    """A proxy for the os.environ which allows for getting/setting typed values."""

    def __init__(self, resolver: Resolver):
        self.resolver = resolver
        for t in STDLIB_TYPES:
            self.register(t)
        for name, t in inspect.getmembers(
            types, lambda o: inspect.isclass(o) and not issubclass(o, Exception)
        ):
            self.register(t, name=name)

    def __getattr__(self, item):
        if inspect.isclass(item):
            t: Type[_ET] = getattr(builtins, item, None) or globals().get(item)
            return self.register(t, name=item)
        raise AttributeError(
            f"{self.__class__.__name__!r} object has no attribute {item!r}"
        )

    def __contains__(self, item):
        return os.environ.__contains__(item)

    def __getitem__(self, item):
        return os.environ.__getitem__(item)

    def __setitem__(self, key, value):
        return self.setenv(key, value)

    def register(self, t: Type[_ET], *aliases: str, name: str = None):
        """Register a handler for the target type `t`."""
        anno = self.resolver.annotation(t)
        if isinstance(
            anno, (common.ForwardDelayedAnnotation, common.DelayedAnnotation)
        ):
            anno = anno.resolved.annotation  # type: ignore

        name = name or get_name(anno.resolved)
        if name in self.__dict__:
            return self.__dict__[name]

        if not inspect.isclass(anno.resolved):
            raise EnvironmentTypeError(
                f"Can't coerce to target {name!r} with t: {t!r}."
            ) from None

        def get(var: str, *, ci: bool = True, default: _ET = ...):  # type: ignore
            return self.getenv(var, default, *aliases, t=t, ci=ci)

        setattr(self, name, get)
        return get

    def getenv(
        self,
        var: str,
        default: _ET = ...,  # type: ignore
        *aliases: str,
        t: Type[_ET] = Any,  # type: ignore
        ci: bool = True,
    ) -> _ET:
        """Get the value of an Environment Variable.

        Keyword Args:
            t: If provided, the type to validate the value against.
            ci: Whether the variable should be considered case-insensitive.
        """
        proto = self.resolver.resolve(t)
        names = {*aliases}
        environ: Mapping[str, str] = os.environ
        if ci:
            var = var.lower()
            names = {v.lower() for v in aliases}
            environ = {k.lower(): value for k, value in os.environ.items()}
        value = environ.get(var, default)
        if value == default and names:
            value = next((environ[k] for k in environ.keys() & names), default)
        if value is ... and t is Any:
            return None  # type: ignore
        if value is ... and not proto.annotation.optional:
            raise EnvironmentValueError(
                f"{var!r} should be of {t!r}, got nothing."
            ) from None
        if value == default:
            return value  # type: ignore
        try:
            return proto.transmute(value)  # type: ignore
        except (TypeError, ValueError, KeyError) as err:
            raise EnvironmentValueError(
                f"Couldn't parse <{var}:{value}> to {t!r}: {err}."
            ) from None

    def setenv(self, var: str, value: Any):
        """Set the `value` as `var` in the OS environ."""
        if isinstance(value, bytes):
            os.environb[var.encode()] = value
            return
        elif not isinstance(value, str):
            value = self.resolver.tojson(value)
        os.environ[var] = value



from __future__ import annotations

import dataclasses
import enum
import warnings
from typing import (
    Union,
    Type,
    Mapping,
    Optional,
    Dict,
    Any,
    List,
    Generic,
    cast,
    AnyStr,
    TYPE_CHECKING,
    MutableMapping,
    Tuple,
)

import inflection  # type: ignore

from typic.common import ReadOnly, WriteOnly
from typic.serde.resolver import resolver
from typic.serde.common import SerdeProtocol, Annotation
from typic.compat import Final, TypedDict, ForwardRef, Literal
from typic.util import get_args, origin, get_name
from typic.checks import istypeddict, isnamedtuple, isliteral, isuniontype
from typic.types.frozendict import FrozenDict

from .field import (  # type: ignore
    MultiSchemaField,
    ObjectSchemaField,
    UndeclaredSchemaField,
    Ref,
    SchemaFieldT,
    SCHEMA_FIELD_FORMATS,
    SchemaType,
    ArraySchemaField,
    NullSchemaField,
)

if TYPE_CHECKING:
    from typic.constraints import ArrayConstraints, MappingConstraints  # noqa: F401

_IGNORE_DOCS = frozenset({Mapping.__doc__, Generic.__doc__, List.__doc__})

__all__ = ("SchemaBuilder", "SchemaDefinitions", "builder")


_KNOWN = (*(f for f in SCHEMA_FIELD_FORMATS.keys() if f not in {AnyStr, object}),)


class SchemaDefinitions(TypedDict):
    """A :py:class:`TypedDict` for JSON Schema Definitions."""

    definitions: Dict[str, Union[ObjectSchemaField, Mapping]]


class SchemaBuilder:
    """Build a JSON schema from an object.

    Notes
    -----
    This shouldn't be used directly.
    Schema generation is handled automatically when a class is wrapped.

    See Also
    --------
    :py:func:`~typic.typed.glob.wrap_cls`
    :py:func:`~typic.klass.klass`
    """

    def __init__(self):
        self.__cache = {}
        self.__attached = set()
        self.__stack = set()

    def attach(self, t: Type):
        self.__attached.add(t)

    def _handle_mapping(
        self, proto: SerdeProtocol, parent: Type = None, *, name: str = None, **extra
    ) -> MutableMapping:
        anno = proto.annotation
        args = anno.args
        config = extra
        config["title"] = self.defname(anno.resolved, name=name)
        doc = getattr(anno.resolved, "__doc__", None)
        if doc not in _IGNORE_DOCS:
            config["description"] = doc

        constraints = cast("MappingConstraints", proto.constraints)
        attrs = (
            ("items", "properties"),
            ("patterns", "patternProperties"),
            ("key_dependencies", "dependencies"),
        )
        for src, target in attrs:
            items = getattr(constraints, src)
            if items:
                config[target] = {
                    nm: self.get_field(
                        resolver.resolve(
                            it.type, is_optional=it.nullable, namespace=parent
                        ),
                        parent=parent,
                    )
                    for nm, it in items.items()
                }
        config["additionalProperties"] = not constraints.total
        if args:
            config["additionalProperties"] = self.get_field(
                resolver.resolve(args[-1], namespace=parent), parent=parent
            )

        return config

    def _handle_array(
        self, proto: SerdeProtocol, parent: Type = None, **extra
    ) -> MutableMapping:
        anno = proto.annotation
        args = anno.args
        has_ellipsis = args[-1] is Ellipsis if args else False
        config = extra
        if has_ellipsis:
            args = args[:-1]
        if args:
            constrs = set(
                self.get_field(resolver.resolve(x, namespace=parent), parent=parent)
                for x in args
            )
            config["items"] = (*constrs,) if len(constrs) > 1 else constrs.pop()
            if anno.origin in {tuple, frozenset}:
                config["additionalItems"] = False if not has_ellipsis else None
            if anno.origin in {set, frozenset}:
                config["uniqueItems"] = True
        return config

    def _check_optional(
        self,
        annotation: Annotation,
        schema: SchemaFieldT,
        ro: Optional[bool],
        wo: Optional[bool],
        name: Optional[str],
    ) -> SchemaFieldT:
        if annotation.optional and annotation.parameter.default is not ...:
            null = NullSchemaField()
            if isinstance(schema, MultiSchemaField):
                anyOf = schema.anyOf or ()
                if null not in {*anyOf}:
                    schema = dataclasses.replace(schema, anyOf=(*anyOf, null))
                return schema
            # Primitives are generally defined as anonymous types
            title = subtitle = schema.title or self.defname(annotation.resolved, name)
            # If we have an object with this name, make this an 'Optional' reference.
            if isinstance(schema, (Ref, ObjectSchemaField)):
                title = subtitle and f"Optional{subtitle}"
            default = getattr(schema, "default", None)
            child: SchemaFieldT = (
                schema
                if isinstance(schema, Ref)
                else dataclasses.replace(
                    schema, default=None, readOnly=ro, writeOnly=wo
                )
            )
            return MultiSchemaField(
                title=title,
                default=default,
                readOnly=ro,
                writeOnly=wo,
                anyOf=(child, NullSchemaField()),
            )
        return schema

    def _handle_union(
        self,
        anno: Annotation,
        ro: Optional[bool],
        wo: Optional[bool],
        name: Optional[str],
        parent: Optional[Type],
    ):
        fields: List[SchemaFieldT] = []
        args = get_args(anno.un_resolved)
        for t in args:
            if t.__class__ is ForwardRef or t is parent:
                n = name or get_name(t)
                fields.append(Ref(f"#/definitions/{n}"))
                continue
            fields.append(
                self.get_field(resolver.resolve(t, namespace=parent), parent=parent)
            )
        schema = self._check_optional(
            anno,
            MultiSchemaField(
                title=name and self.defname(anno.resolved, name=name),
                anyOf=(*fields,),
            ),
            ro,
            wo,
            name,
        )
        self.__cache[anno] = schema
        return schema

    def _build_field(
        self,
        use: Type,
        protocol: SerdeProtocol,
        parent: Optional[Type],
        enum_: Optional[Tuple[Any, ...]],
        default: Optional[Any],
        ro: Optional[bool],
        wo: Optional[bool],
        name: Optional[str],
    ) -> SchemaFieldT:
        # If we've got a base object, use it
        base: Optional[SchemaFieldT]
        if use is object:
            base = UndeclaredSchemaField()
        elif istypeddict(use) or isnamedtuple(use):
            base = None
        else:
            base = cast(SchemaFieldT, SCHEMA_FIELD_FORMATS.get_by_parent(use))
        if base:
            config: MutableMapping = (
                protocol.constraints.for_schema() if protocol.constraints else {}
            )
            config.update(enum=enum_, default=default, readOnly=ro, writeOnly=wo)
            # `use` should always be a dict if the annotation is a Mapping,
            # thanks to `origin()` & `resolve()`.
            if isinstance(base, ObjectSchemaField):
                config = self._handle_mapping(
                    protocol, parent=parent, name=name, **config
                )
            elif isinstance(base, ArraySchemaField):
                config = self._handle_array(protocol, parent=parent, **config)
            schema = dataclasses.replace(base, **config)
        else:
            try:
                schema = self.build_schema(use)
            except (ValueError, TypeError) as e:
                warnings.warn(f"Couldn't build schema for {use}: {e}")
                schema = UndeclaredSchemaField(
                    enum=enum_,
                    title=self.defname(use, name=name),
                    default=default,
                    readOnly=ro,
                    writeOnly=wo,
                )
        schema = self._check_optional(protocol.annotation, schema, ro, wo, name)
        return schema

    def get_field(
        self,
        protocol: SerdeProtocol,
        *,
        ro: bool = None,
        wo: bool = None,
        name: str = None,
        parent: Type = None,
    ) -> "SchemaFieldT":
        """Get a field definition for a JSON Schema."""
        if protocol.annotation in self.__stack:
            name = self.defname(protocol.annotation.resolved_origin, name)
            return self._check_optional(
                protocol.annotation, Ref(f"#/definitions/{name}"), ro, wo, name
            )
        anno = protocol.annotation
        if anno in self.__cache:
            return self.__cache[anno]
        # Get the default value
        # `None` gets filtered out down the line. this is okay.
        # If a field isn't required an empty default is functionally the same
        # as a default to None for the JSON schema.
        default = anno.parameter.default if anno.has_default else None
        # `use` is the based annotation we will use for building the schema
        use = getattr(anno.origin, "__parent__", anno.origin)
        # This is a flat optional, handle it separately from the Union block.
        use = anno.resolved if isuniontype(use) and not anno.args else use
        # If there's not a static annotation, short-circuit the rest of the checks.
        schema: SchemaFieldT
        if use in {Any, anno.EMPTY}:
            schema = self._check_optional(anno, UndeclaredSchemaField(), ro, wo, name)
            self.__cache[anno] = schema
            return schema

        # Unions are `anyOf`, get a new field for each arg and return.
        # {'type': ['string', 'integer']} ==
        #   {'anyOf': [{'type': 'string'}, {'type': 'integer'}]}
        # We don't care about syntactic sugar if it's functionally the same.
        if isuniontype(use):
            return self._handle_union(anno=anno, ro=ro, wo=wo, name=name, parent=parent)

        self.__stack.add(anno)

        # Check if this should be ro/wo
        if use in {ReadOnly, WriteOnly, Final}:
            ro = (use in {ReadOnly, Final}) or None
            wo = (use is WriteOnly) or None
            use = origin(anno.resolved)
            use = getattr(use, "__parent__", use)

        # Check for an enumeration
        enum_ = None
        # Functionally, literals are enumerations.
        if isliteral(use):
            enum_ = (*(a for a in anno.args if a is not None),)
            ts = {a.__class__ for a in enum_}
            use = Literal
            if len(ts) == 1:
                use = ts.pop()

        elif issubclass(use, enum.Enum):
            use = cast(Type[enum.Enum], use)
            enum_ = tuple(x.value for x in use)
            use = getattr(use._member_type_, "__parent__", use._member_type_)  # type: ignore

        # If this is ro with a default, we can consider this a const
        # Which is an enum with a single value -
        # we don't currently honor `{'const': <val>}` since it's just syntactic sugar.
        if ro and default:
            enum_ = (default.value if isinstance(default, enum.Enum) else default,)

        schema = self._build_field(
            use=use,
            protocol=protocol,
            parent=parent,
            enum_=enum_,
            default=default,
            ro=ro,
            wo=wo,
            name=name,
        )
        self.__cache[anno] = schema
        self.__stack.clear()
        return schema

    @staticmethod
    def _flatten_object_definitions(
        definitions: Dict[str, Any], field: ObjectSchemaField
    ):
        definitions.update(**(field.definitions or {}))  # type: ignore
        field = dataclasses.replace(field, definitions=None)
        definitions[field.title] = field  # type: ignore
        return Ref(f"#/definitions/{field.title}")

    def _flatten_array_definitions(
        self, definitions: Dict[str, Any], field: ArraySchemaField
    ):
        replace = {}
        for field_name in ("items", "contains", "additionalItems"):
            it = getattr(field, field_name)
            if isinstance(it, (ObjectSchemaField, ArraySchemaField, MultiSchemaField)):
                ref = self._flatten_definitions(definitions, it)
                replace[field_name] = ref
        return dataclasses.replace(field, **replace) if replace else field

    def _flatten_multi_definitions(
        self, definitions: Dict[str, Any], field: MultiSchemaField
    ) -> MultiSchemaField:
        replace = {}
        for field_name in ("anyOf", "allOf", "oneOf"):
            it = getattr(field, field_name)
            if it:
                flattened = []
                for f in it:
                    nf = self._flatten_definitions(definitions, f)
                    flattened.append(nf)
                replace[field_name] = (*flattened,)
        return dataclasses.replace(field, **replace) if replace else field

    def _flatten_definitions(self, definitions: Dict[str, Any], field: SchemaFieldT):
        if isinstance(field, ObjectSchemaField):
            return self._flatten_object_definitions(definitions, field)
        elif isinstance(field, ArraySchemaField):
            return self._flatten_array_definitions(definitions, field)
        elif isinstance(field, MultiSchemaField):
            return self._flatten_multi_definitions(definitions, field)
        return field

    @classmethod
    def defname(cls, obj, name: str = None) -> Optional[str]:
        """Get the definition name for an object."""
        defname = name or getattr(obj, "__name__", None)
        if defname in cls._IGNORE_NAME:
            defname = None
        if (obj is dict or origin(obj) is dict) and name:
            defname = name
        return inflection.camelize(defname) if defname else None

    _IGNORE_NAME = {"Mapping", "MutableMapping", "Dict"}

    def build_schema(self, obj: Type, *, name: str = None) -> ObjectSchemaField:
        """Build a valid JSON Schema, including nested schemas."""
        if obj in self.__cache:  # pragma: nocover
            return self.__cache[obj]

        protocols: Dict[str, SerdeProtocol] = resolver.protocols(obj)
        definitions: Dict[str, Any] = {}
        properties: Dict[str, Any] = {}
        required: List[str] = []
        total: bool = getattr(obj, "__total__", True)
        for nm, protocol in protocols.items():
            if protocol.annotation.resolved_origin is obj:
                flattened = Ref(f"#/definitions/{self.defname(obj)}")
            else:
                ro = protocol.annotation.is_class_var or None
                field = self.get_field(protocol, name=nm, parent=obj, ro=ro)
                # If we received an object schema,
                # figure out a name and inherit the definitions.
                flattened = self._flatten_definitions(definitions, field)
            properties[nm] = flattened
            # Check for required field(s)
            if not protocol.annotation.has_default:
                required.append(nm)
        schema = ObjectSchemaField(
            title=name or self.defname(obj),
            description=obj.__doc__,
            properties=FrozenDict(properties),
            additionalProperties=False,
            required=(*required,) if total else (),
            definitions=FrozenDict(definitions),
        )
        self.__cache[obj] = schema
        return schema

    def all(self, primitive: bool = False) -> SchemaDefinitions:
        """Get all of the JSON Schema objects which have been defined.

        `typical` maintains a register of all resolved schema definitions.
        This method give high-level, immediate read access to that registry.

        Parameters
        ----------
        primitive

        Examples
        --------
        >>> import json
        >>> import typic
        >>> schemas = typic.schemas(primitive=True)
        >>> print(json.dumps(schemas["definitions"]["Duck"], indent=2))
        {
          "type": "object",
          "title": "Duck",
          "description": "Duck(color: str)",
          "properties": {
            "color": {
              "type": "string"
            }
          },
          "additionalProperties": false,
          "required": [
            "color"
          ]
        }
        """
        definitions = SchemaDefinitions(definitions={})
        schm: ObjectSchemaField
        while self.__attached:
            self.build_schema(self.__attached.pop())
        for obj, schm in self.__cache.items():
            if schm.type != SchemaType.OBJ:
                continue
            definitions["definitions"].update(
                {
                    x: dataclasses.replace(y, definitions=None)
                    for x, y in (schm.definitions or {}).items()
                }
            )
            definitions["definitions"][
                schm.title  # type: ignore
            ] = dataclasses.replace(schm, definitions=None)
        if primitive:
            definitions["definitions"] = {
                x: y.primitive() if isinstance(y, ObjectSchemaField) else y
                for x, y in definitions["definitions"].items()
            }
        return definitions


builder = SchemaBuilder()



from __future__ import annotations

import dataclasses
import inspect
from collections import deque
from typing import (
    Dict,
    Deque,
    Any,
    Tuple,
    TYPE_CHECKING,
    List,
    Optional,
    Union,
    Type,
    Callable,
    Mapping,
    Iterable,
)

from typic import util
from ..common import (
    EMPTY,
    POSITIONAL_ONLY,
    RETURN_KEY,
    TOO_MANY_POS,
    VAR_POSITIONAL,
    VAR_KEYWORD,
    KWD_KINDS,
)

if TYPE_CHECKING:  # pragma: nocover
    from .resolver import Resolver  # noqa: F401
    from .common import SerdeProtocol, SerdeProtocolsT  # noqa: F401


@dataclasses.dataclass(frozen=True)
class BoundArguments:
    obj: Union[Type, Callable]
    """The object we "bound" the input to."""
    annotations: SerdeProtocolsT
    """A mapping of the resolved annotations."""
    parameters: Mapping[str, inspect.Parameter]
    """A mapping of the parameters."""
    arguments: Dict[str, Any]
    """A mapping of the input to parameter name."""
    returns: Optional["SerdeProtocol"]
    """The resolved return type, if any."""
    _argnames: Tuple[str, ...]
    _kwdargnames: Tuple[str, ...]

    @util.cached_property
    def args(self) -> Tuple[Any, ...]:
        """A tuple of the args passed to the callable."""
        args: List = list()
        argsappend = args.append
        argsextend = args.extend
        paramsget = self.parameters.__getitem__
        argumentsget = self.arguments.__getitem__
        for name in self._argnames:
            kind = paramsget(name).kind
            arg = argumentsget(name)
            if kind == VAR_POSITIONAL:
                argsextend(arg)
            else:
                argsappend(arg)
        return tuple(args)

    @util.cached_property
    def kwargs(self) -> Dict[str, Any]:
        """A mapping of the key-word arguments passed to the callable."""
        kwargs: Dict = {}
        kwargsupdate = kwargs.update
        kwargsset = kwargs.__setitem__
        paramsget = self.parameters.__getitem__
        argumentsget = self.arguments.__getitem__
        for name in self._kwdargnames:
            kind = paramsget(name).kind
            arg = argumentsget(name)
            if kind == VAR_KEYWORD:
                kwargsupdate(arg)
            else:
                kwargsset(name, arg)
        return kwargs

    def eval(self) -> Any:
        """Evaluate the callable against the input provided.

        Examples
        --------
        >>> import typic
        >>>
        >>> def foo(bar: int) -> int:
        ...     return bar ** bar
        ...
        >>> bound = typic.bind(foo, "2")
        >>> bound.eval()
        4
        """
        return self.obj(*self.args, **self.kwargs)


class Binder:
    def __init__(self, resolver: Resolver):
        self.resolver = resolver

    def _bind_posargs(
        self,
        arguments: Dict[str, Any],
        params: Deque[inspect.Parameter],
        annos: "SerdeProtocolsT",
        args: Deque[Any],
        kwargs: Dict[str, Any],
    ) -> Tuple[str, ...]:
        # Bind any positional arguments

        # bytecode hack to localize access
        # only noticeable with really large datasets
        # but it's best to be prepared.
        posargs: List[str] = []
        posargsadd = posargs.append
        argspop = args.popleft
        paramspop = params.popleft
        annosget = annos.get
        argumentsset = arguments.__setitem__
        while args and params:
            val = argspop()
            param: inspect.Parameter = paramspop()
            name = param.name
            anno: Optional[SerdeProtocol] = annosget(name)
            kind = param.kind
            # We've got varargs, so push all supplied args to that param.
            if kind == VAR_POSITIONAL:
                value = (val, *args)
                args = deque()
                if anno:
                    value = anno(value)
                argumentsset(name, value)
                posargsadd(name)
                break

            # We're not supposed to have kwdargs....
            if kind in KWD_KINDS:
                raise TypeError(TOO_MANY_POS) from None

            # Passed in by ref and assignment... no good.
            if name in kwargs:
                raise TypeError(f"multiple values for argument '{name}'") from None

            # We're g2g
            value = anno(val) if anno else val
            argumentsset(name, value)
            posargsadd(name)

        if args:
            raise TypeError(TOO_MANY_POS) from None

        return tuple(posargs)

    def _bind_kwdargs(
        self,
        arguments: Dict[str, Any],
        params: Deque[inspect.Parameter],
        annos: Dict[str, SerdeProtocol],
        kwargs: Dict[str, Any],
        partial: bool = False,
    ) -> Tuple[str, ...]:
        # Bind any key-word arguments
        kwdargs: List[str] = list()
        kwdargsadd = kwdargs.append
        kwargs_anno = None
        kwdargs_param = None
        kwargspop = kwargs.pop
        annosget = annos.get
        argumentsset = arguments.__setitem__
        for param in params:
            kind = param.kind
            name = param.name
            anno = annosget(name)
            # Move on, but don't forget
            if kind == VAR_KEYWORD:
                kwargs_anno = anno
                kwdargs_param = param
                continue
            # We don't care about these
            if kind == VAR_POSITIONAL:
                continue
            # try to bind the parameter
            if name in kwargs:
                val = kwargspop(name)
                if kind == POSITIONAL_ONLY:
                    raise TypeError(
                        f"{name!r} parameter is positional only,"
                        "but was passed as a keyword."
                    )
                value = anno(val) if anno else val
                argumentsset(name, value)
                kwdargsadd(name)
            elif not partial and param.default is EMPTY:
                raise TypeError(f"missing required argument: {name!r}")

        # We didn't clear out all the kwdargs. Check to see if we came across a **kwargs
        if kwargs:
            if kwdargs_param is not None:
                # Process our '**kwargs'-like parameter
                name = kwdargs_param.name
                value = kwargs_anno.transmute(kwargs) if kwargs_anno else kwargs  # type: ignore
                argumentsset(name, value)
                kwdargsadd(name)
            else:
                raise TypeError(
                    f"'got an unexpected keyword argument {next(iter(kwargs))!r}'"
                )

        return tuple(kwdargs)

    def _bind_input(
        self,
        obj: Union[Type, Callable],
        annos: "SerdeProtocolsT",
        params: Mapping[str, inspect.Parameter],
        args: Iterable[Any],
        kwargs: Dict[str, Any],
        *,
        partial: bool = False,
    ) -> BoundArguments:
        """Bind annotations and parameters to received input.

        Taken approximately from :py:meth:`inspect.Signature.bind`, with a few changes.

        About 10% faster, on average, and coerces values with their annotation if possible.

        Parameters
        ----------
        annos
            A mapping of :py:class:`ResolvedAnnotation` to param name.
        params
            A mapping of :py:class:`inspect.Parameter` to param name.
        args
            The positional args to bind to their param and annotation, if possible.
        kwargs
            The keyword args to bind to their param and annotation, if possible.

        Other Parameters
        ----------------
        partial
            Bind a partial input.

        Raises
        ------
        TypeError
            If we can't match up the received input to the signature
        """
        arguments: Dict[str, Any] = dict()
        returns = annos.pop(RETURN_KEY, None)
        args = deque(args)
        parameters = deque(params.values())
        # Bind any positional arguments.
        posargs = self._bind_posargs(arguments, parameters, annos, args, kwargs)
        # Bind any keyword arguments.
        kwdargs = self._bind_kwdargs(arguments, parameters, annos, kwargs, partial)
        return BoundArguments(obj, annos, params, arguments, returns, posargs, kwdargs)

    def bind(
        self,
        obj: Union[Type, Callable],
        *args: Any,
        partial: bool = False,
        coerce: bool = True,
        strict: bool = False,
        **kwargs: Mapping[str, Any],
    ) -> BoundArguments:
        """Bind a received input to a callable or object's signature.

        If we can locate an annotation for any args or kwargs, we'll automatically
        coerce as well.

        This implementation is similar to :py:meth:`inspect.Signature.bind`,
        but is ~10-20% faster.
        We also use a cached the signature to avoid the expense of that call if possible.

        Parameters
        ----------
        obj
            The object you wish to bind your input to.
        *args
            The given positional args.
        partial
            Whether to bind a partial input.
        coerce
            Whether to coerce the input to the annotation provided.
        strict
            Whether to validate the input against the annotation provided.
        **kwargs
            The given keyword args.

        Returns
        -------
        :py:class:`BoundArguments`
            The bound and coerced arguments.

        Raises
        ------
        TypeError
            If we can't match up the received input to the signature

        Examples
        --------
        >>> import typic
        >>>
        >>> def add(a: int, b: int, *, c: int = None) -> int:
        ...     return a + b + (c or 0)
        ...
        >>> bound = typic.bind(add, "1", "2", c=3.0)
        >>> bound.arguments
        {'a': 1, 'b': 2, 'c': 3}
        >>> bound.args
        (1, 2)
        >>> bound.kwargs
        {'c': 3}
        >>> bound.eval()
        6
        >>> typic.bind(add, 1, 3.0, strict=True)
        Traceback (most recent call last):
            ...
        typic.constraints.error.ConstraintValueError: Given value <3.0> fails constraints: (type=int, nullable=False)
        """
        return self._bind_input(
            obj=obj,
            annos=self.resolver.protocols(obj, strict=strict)
            if (coerce or strict)
            else {},
            params=util.cached_signature(obj).parameters,
            args=args,
            kwargs=kwargs,
            partial=partial,
        )



from __future__ import annotations

import ast
import collections
import collections.abc
import dataclasses
import functools
import inspect
import sys
import types
import warnings
from datetime import date, datetime, timedelta, time
from threading import RLock
from types import MappingProxyType, MemberDescriptorType
import typing
from typing import (  # type: ignore  # ironic...
    Tuple,
    Any,
    Sequence,
    Collection,
    Mapping,
    Hashable,
    Type,
    TypeVar,
    Callable,
    get_type_hints as _get_type_hints,
    Union,
    MutableMapping,
    MutableSequence,
    Iterable,
    AbstractSet,
    MutableSet,
    Dict,
    Optional,
    _eval_type,
)

import pendulum
from future_typing import transform_annotation

import typic.checks as checks
from typic.compat import ForwardRef, lru_cache, KW_ONLY, get_origin
from typic.ext import json

__all__ = (
    "cached_issubclass",
    "cached_property",
    "cached_signature",
    "cached_simple_attributes",
    "cached_type_hints",
    "cachedmethod",
    "extract",
    "fastcachedmethod",
    "filtered_repr",
    "get_args",
    "get_name",
    "get_defname",
    "get_qualname",
    "get_tag_for_types",
    "get_type_hints",
    "get_unique_name",
    "isoformat",
    "origin",
    "resolve_supertype",
    "safe_eval",
    "safe_get_params",
    "signature",
    "simple_attributes",
    "slotted",
    "TaggedUnion",
    "typed_dict_signature",
    "TypeMap",
)

from typic.compat import SQLAMetaData, sqla_registry


GENERIC_TYPE_MAP = {
    Sequence: list,
    MutableSequence: list,
    collections.abc.Sequence: list,
    collections.abc.MutableSequence: list,
    Collection: list,
    collections.abc.Collection: list,
    Iterable: list,
    collections.abc.Iterable: list,
    AbstractSet: set,
    MutableSet: set,
    collections.abc.Set: set,
    collections.abc.MutableSet: set,
    Mapping: dict,
    MutableMapping: dict,
    collections.abc.Mapping: dict,
    collections.abc.MutableMapping: dict,
    Hashable: str,
    collections.abc.Hashable: str,
}


@lru_cache(maxsize=2000, typed=True)
def safe_eval(string: str) -> Tuple[bool, Any]:
    """Try a few methods to evaluate a string and get the correct Python data-type.

    Return the result and an indicator for whether we could do anything with it.

    Examples
    --------
    >>> safe_eval('{"foo": null}')
    (True, {'foo': None})

    Parameters
    ----------
    string
        The string to attempt to evaluate into a valid Python data-structure or object

    Returns
    -------
    processed :
        Whether we successfully evaluated the string
    result :
        The final result of the operation
    """
    try:
        result, processed = ast.literal_eval(string), True
    except (TypeError, ValueError, SyntaxError):
        try:
            result, processed = json.loads(string), True
        except (TypeError, ValueError, SyntaxError):
            result, processed = string, False

    return processed, result


def _check_generics(hint: Any):
    return GENERIC_TYPE_MAP.get(hint, hint)


def filtered_repr(self) -> str:
    return f"{type(self).__name__}{filtered_str(self)}"


def filtered_str(self) -> str:
    fields = []
    for f in dataclasses.fields(self):
        val = getattr(self, f.name)
        if (val or val in {False, 0}) and f.repr:
            fields.append(f"{f.name}={val!r}")
    return f"({', '.join(fields)})"


@lru_cache(maxsize=None)
def origin(annotation: Any) -> Any:
    """Get the highest-order 'origin'-type for subclasses of typing._SpecialForm.

    For the purposes of this library, if we can resolve to a builtin type, we will.

    Examples
    --------
    >>> import typic
    >>> from typing import Dict, Mapping, NewType, Optional
    >>> typic.origin(Dict)
    <class 'dict'>
    >>> typic.origin(Mapping)
    <class 'dict'>
    >>> Registry = NewType('Registry', Dict)
    >>> typic.origin(Registry)
    <class 'dict'>
    >>> class Foo: ...
    ...
    >>> typic.origin(Foo)
    <class 'typic.util.Foo'>
    """
    # Resolve custom NewTypes.
    actual = resolve_supertype(annotation)

    # Unwrap optional/classvar
    if checks.isclassvartype(actual):
        args = get_args(actual)
        actual = args[0] if args else actual

    actual = get_origin(actual) or actual

    # provide defaults for generics
    if not checks.isbuiltintype(actual):
        actual = _check_generics(actual)

    if inspect.isroutine(actual):
        actual = collections.abc.Callable

    return actual


@lru_cache(maxsize=None)
def get_args(annotation: Any) -> Tuple[Any, ...]:
    """Get the args supplied to an annotation, excluding :py:class:`typing.TypeVar`.

    Examples
    --------
    >>> import typic
    >>> from typing import Dict, TypeVar
    >>> T = TypeVar("T")
    >>> typic.get_args(Dict)
    ()
    >>> typic.get_args(Dict[str, int])
    (<class 'str'>, <class 'int'>)
    >>> typic.get_args(Dict[str, T])
    (<class 'str'>,)
    """
    return (
        *(x for x in getattr(annotation, "__args__", ()) if type(x) is not TypeVar),
    )


@lru_cache(maxsize=None)
def get_name(obj: Union[Type, ForwardRef, Callable]) -> str:
    """Safely retrieve the name of either a standard object or a type annotation.

    Examples
    --------
    >>> import typic
    >>> from typing import Dict, Any
    >>> T = TypeVar("T")
    >>> typic.get_name(Dict)
    'Dict'
    >>> typic.get_name(Dict[str, str])
    'Dict'
    >>> typic.get_name(Any)
    'Any'
    >>> typic.get_name(dict)
    'dict'
    """
    strobj = get_qualname(obj)
    return strobj.rsplit(".")[-1]


@lru_cache(maxsize=None)
def get_qualname(obj: Union[Type, ForwardRef, Callable]) -> str:
    """Safely retrieve the qualname of either a standard object or a type annotation.

    Examples
    --------
    >>> import typic
    >>> from typing import Dict, Any
    >>> T = TypeVar("T")
    >>> typic.get_qualname(Dict)
    'typing.Dict'
    >>> typic.get_qualname(Dict[str, str])
    'typing.Dict'
    >>> typic.get_qualname(Any)
    'typing.Any'
    >>> typic.get_qualname(dict)
    'dict'
    """
    strobj = str(obj)
    isgeneric = strobj.startswith("typing.")
    if not isgeneric and isinstance(obj, ForwardRef):
        strobj = str(obj.__forward_arg__)
        isgeneric = strobj.startswith("typing.")
    # We got a typing thing.
    if isgeneric:
        # If this is a subscripted generic we should clean that up.
        return strobj.split("[")[0]
    # Easy-ish path, use name magix
    if hasattr(obj, "__qualname__") and obj.__qualname__:  # type: ignore
        qualname = obj.__qualname__  # type: ignore
        if "<locals>" in qualname:
            return qualname.rsplit(".")[-1]
        return qualname
    if hasattr(obj, "__name__") and obj.__name__:  # type: ignore
        return obj.__name__  # type: ignore
    return strobj


@lru_cache(maxsize=None)
def get_unique_name(obj: Type) -> str:
    return f"{get_name(obj)}_{id(obj)}".replace("-", "_")


@lru_cache(maxsize=None)
def get_defname(pre: str, obj: Hashable) -> str:
    return f"{pre}_{hash(obj)}".replace("-", "_")


@lru_cache(maxsize=None)
def resolve_supertype(annotation: Type[Any]) -> Any:
    """Get the highest-order supertype for a NewType.

    Examples
    --------
    >>> import typic
    >>> from typing import NewType
    >>> UserID = NewType("UserID", int)
    >>> AdminID = NewType("AdminID", UserID)
    >>> typic.resolve_supertype(AdminID)
    <class 'int'>
    """
    while hasattr(annotation, "__supertype__"):
        annotation = annotation.__supertype__
    return annotation


class cached_property:  # type: ignore
    def __init__(self, func):
        self.func = func
        self.__doc__ = func.__doc__
        self.lock = RLock()

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        attrname = self.func.__name__
        try:
            cache = instance.__dict__
        # objects with __slots__ have no __dict__
        except AttributeError:  # pragma: nocover
            msg = (
                f"No '__dict__' attribute on {type(instance).__name__!r} "
                f"instance to cache {attrname!r} property."
            )
            raise TypeError(msg) from None
        with self.lock:
            # check if another thread filled cache while we awaited lock
            if attrname not in cache:
                cache[attrname] = self.func(instance)
        return cache[attrname]


# stolen from functools._HashedSeq
class __HashedSeq(list):  # pragma: nocover

    __slots__ = "hashvalue"

    def __init__(self, tup, hash=hash):
        self[:] = tup
        self.hashvalue = hash(tup)

    def __hash__(self):
        return self.hashvalue


# Stolen approximately from functools._make_key
# Should try to find a faster way to do this.
def _make_key(
    args,
    kwds,
    kwd_mark=(object(),),
    fasttypes=frozenset({int, str}),
    type=type,
    len=len,
):  # pragma: nocover
    key = args
    if kwds:
        key += kwd_mark
        for item in kwds.items():
            key += item
    elif len(key) == 1 and type(key[0]) in fasttypes:
        return key[0]
    return __HashedSeq(key)


_T = TypeVar("_T")


def cachedmethod(func: Callable[..., _T]) -> Callable[..., _T]:  # pragma: nocover
    """Thread-safe caching of the result of an instance method.

    Mimics some of the pure-Python implementation of :py:func:`functools.lru_cache`.
    Major differences are that it's un-bounded in size and no direct access to the cache.
    It's also slower, being that it's pure-Python and the actual `_lru_cache_wrapper`
    is implemented in C.
    """

    cache = func.__dict__
    cacheget = cache.get
    makekey = _make_key
    sentinel = object()
    lock = RLock()

    @functools.wraps(func)
    def _cached_method_wrapper(*args, **kwargs) -> _T:
        nonlocal cache
        key = makekey(args[1:], kwargs)
        result = cacheget(key, sentinel)
        if result is not sentinel:
            return result
        with lock:
            result = func(*args, **kwargs)
            cache[key] = result
        return result

    return _cached_method_wrapper


def fastcachedmethod(func):
    """Super-fast memoization of a method.

    Notes
    -----
    This is limited to a method that takes only one arg and is un-bounded in size.

    Examples
    --------
    >>> import typic
    >>>
    >>> class Foo:
    ...     @typic.fastcachedmethod
    ...     def bar(self, num: int) -> int:
    ...         return num * (num + 1)
    ...
    """

    class memodict(dict):
        __slots__ = ()

        def __missing__(self, key):
            self[key] = ret = func(instance, key)
            return ret

    sentinel = object()
    memo = memodict()
    memoget = memo.__getitem__
    instance = sentinel

    @functools.wraps(func)
    def _fast_cached_method_wrapper(inst, arg):
        nonlocal instance
        instance = inst
        return memoget(arg)

    _fast_cached_method_wrapper.cache_clear = memo.clear
    _fast_cached_method_wrapper.cache_size = memo.__len__
    _fast_cached_method_wrapper.cache_view = lambda: MappingProxyType(memo)
    _fast_cached_method_wrapper._cache = memo

    return _fast_cached_method_wrapper


def signature(obj: Union[Callable, Type]) -> inspect.Signature:
    """Get the signature of a type or callable.

    Also supports TypedDict subclasses
    """
    return (
        typed_dict_signature(obj)
        if checks.istypeddict(obj)  # type: ignore
        else inspect.signature(obj)
    )


cached_signature = lru_cache(maxsize=None)(signature)


def _safe_get_type_hints(annotation: Union[Type, Callable]) -> Dict[str, Type[Any]]:
    raw_annotations: Dict[str, Any] = {}
    base_globals: Dict[str, Any] = {"typing": typing}
    if isinstance(annotation, type):
        for base in reversed(annotation.__mro__):
            base_globals.update(sys.modules[base.__module__].__dict__)
            raw_annotations.update(getattr(base, "__annotations__", None) or {})
    else:
        raw_annotations = getattr(annotation, "__annotations__", None) or {}
        module_name = getattr(annotation, "__module__", None)
        if module_name:
            base_globals.update(sys.modules[module_name].__dict__)
    annotations = {}
    for name, value in raw_annotations.items():
        if isinstance(value, str):
            value = transform_annotation(value)
            if sys.version_info >= (3, 7):
                value = ForwardRef(value, is_argument=False)
            else:
                value = ForwardRef(value)
        try:
            value = _eval_type(value, base_globals or None, None)
        except NameError:
            # this is ok, we deal with it later.
            pass
        except TypeError as e:
            warnings.warn(f"Couldn't evaluate type {value!r}: {e}")
            value = Any
        annotations[name] = value
    return annotations


def get_type_hints(obj: Union[Type, Callable]) -> Dict[str, Type[Any]]:
    try:
        hints = _get_type_hints(obj)
    except (NameError, TypeError):
        hints = _safe_get_type_hints(obj)
    # KW_ONLY is a special sentinel to denote kw-only params in a dataclass.
    #  We don't want to do anything with this hint/field. It's not real.
    return {f: t for f, t in hints.items() if t is not KW_ONLY}


cached_type_hints = lru_cache(maxsize=None)(get_type_hints)


@lru_cache(maxsize=None)
def cached_issubclass(st: Type, t: Union[Type, Tuple[Type, ...]]) -> bool:
    """A cached result of :py:func:`issubclass`."""
    return issubclass(st, t)


def simple_attributes(t: Type) -> Tuple[str, ...]:
    """Extract all public, static data-attributes for a given type."""
    # If slots are defined, this is the best way to locate static attributes.
    if hasattr(t, "__slots__") and t.__slots__:
        return (
            *(
                f
                for f in t.__slots__
                if not f.startswith("_")
                # JIC - check if this is something fancy.
                and not isinstance(getattr(t, f, ...), _DYNAMIC_ATTRIBUTES)
            ),
        )
    # Otherwise we have to guess. This is inherently faulty, as attributes aren't
    #   always defined on a class before instantiation. The alternative is reverse
    #   engineering the constructor... yikes.
    return (
        *(
            x
            for x, y in inspect.getmembers(t, predicate=checks.issimpleattribute)
            if not x.startswith("_") and not isinstance(y, _DYNAMIC_ATTRIBUTES)
        ),
    )


_DYNAMIC_ATTRIBUTES = (SQLAMetaData, sqla_registry)


cached_simple_attributes = lru_cache(maxsize=None)(simple_attributes)
"""A cached result of :py:func:`simple_attributes`."""


def typed_dict_signature(obj: Callable) -> inspect.Signature:
    """A little faker for getting the "signature" of a :py:class:`TypedDict`.

    Technically, these are dicts at runtime, but we are enforcing a static shape,
    so we should be able to declare a matching signature for it.
    """
    hints = cached_type_hints(obj)
    return inspect.Signature(
        parameters=tuple(
            inspect.Parameter(
                name=x,
                kind=inspect.Parameter.KEYWORD_ONLY,
                annotation=y,
                default=getattr(obj, x, inspect.Parameter.empty),
            )
            for x, y in hints.items()
        )
    )


@lru_cache(maxsize=None)
def safe_get_params(obj: Type) -> Mapping[str, inspect.Parameter]:
    params: Mapping[str, inspect.Parameter]
    try:
        if checks.issubclass(obj, Mapping) and not checks.istypeddict(obj):
            return {}
        params = cached_signature(obj).parameters
    except (ValueError, TypeError):  # pragma: nocover
        params = {}
    return params


VT = TypeVar("VT")


class TypeMap(Dict[Type, VT]):
    """A mapping of Type -> value."""

    def get_by_parent(self, t: Type, default: VT = None) -> Optional[VT]:
        """Traverse the MRO of a class, return the value for the nearest parent."""
        # Skip traversal if this type is already mapped.
        if t in self:
            return self[t]

        # Get the MRO - the first value is the given type so skip it
        try:
            for ptype in inspect.getmro(t)[1:]:
                if ptype in self:
                    v = self[ptype]
                    # Cache for later use
                    self[t] = v
                    return v
        except (AttributeError, TypeError):
            pass

        return default


def slotted(
    _cls: Type = None,
    *,
    dict: bool = True,
    weakref: bool = False,
):
    """Decorator to create a "slotted" version of the provided class.

    Returns new class object as it's not possible to add __slots__ after class creation.

    Source: https://github.com/starhel/dataslots/blob/master/dataslots/__init__.py
    """

    def _slots_setstate(self, state):
        for param_dict in filter(None, state):
            for slot, value in param_dict.items():
                object.__setattr__(self, slot, value)

    def wrap(cls):
        key = repr(cls)
        if key in _stack:
            raise TypeError(
                f"{cls!r} uses a custom metaclass {cls.__class__!r} "
                "which is not compatible with automatic slots. "
                "See Issue #104 on GitHub for more information."
            ) from None

        _stack.add(key)

        cls_dict = {**cls.__dict__}
        # Create only missing slots
        inherited_slots = set().union(
            *(getattr(c, "__slots__", set()) for c in cls.mro())
        )

        field_names = {f.name for f in dataclasses.fields(cls)}
        if dict:
            field_names.add("__dict__")
        if weakref:
            field_names.add("__weakref__")
        cls_dict["__slots__"] = (*(field_names - inherited_slots),)

        # Erase filed names from class __dict__
        for f in field_names:
            cls_dict.pop(f, None)

        # Erase __dict__ and __weakref__
        cls_dict.pop("__dict__", None)
        cls_dict.pop("__weakref__", None)

        # Pickle fix for frozen dataclass as mentioned in https://bugs.python.org/issue36424
        # Use only if __getstate__ and __setstate__ are not declared and frozen=True
        if (
            all(param not in cls_dict for param in ["__getstate__", "__setstate__"])
            and cls.__dataclass_params__.frozen
        ):
            cls_dict["__setstate__"] = _slots_setstate

        # Prepare new class with slots
        new_cls = cls.__class__(cls.__name__, cls.__bases__, cls_dict)
        new_cls.__qualname__ = cls.__qualname__
        new_cls.__module__ = cls.__module__

        _stack.clear()
        return new_cls

    return wrap if _cls is None else wrap(_cls)


_stack: MutableSet[Type] = set()


class joinedrepr(str):
    __slots__ = ("fields", "__dict__")
    fields: Iterable[Any]

    def __new__(cls, *fields):
        n = str.__new__(cls)
        n.fields = fields
        return n

    @cached_property
    def __repr(self) -> str:
        return ".".join(str(f) for f in self.fields)

    def __repr__(self) -> str:
        return self.__repr

    def __str__(self) -> str:
        return self.__repr


class collectionrepr(str):
    __slots__ = (
        "root_name",
        "keys",
        "__dict__",
    )
    root_name: str
    keys: Iterable[Any]

    def __new__(cls, root_name: "ReprT", *keys):
        n = str.__new__(cls)
        n.root_name = root_name
        n.keys = keys
        return n

    @cached_property
    def __repr(self) -> str:
        keys = "".join(f"[{o!r}]" for o in self.keys)
        return f"{self.root_name}{keys}"

    def __repr__(self) -> str:
        return self.__repr

    def __str__(self) -> str:
        return self.__repr


ReprT = Union[str, joinedrepr, collectionrepr]


@functools.lru_cache(maxsize=100_000)
def isoformat(t: Union[date, datetime, time, timedelta]) -> str:
    if isinstance(t, (date, datetime, time)):
        return t.isoformat()
    d = t
    if not isinstance(d, pendulum.Duration):
        d = pendulum.duration(
            days=t.days,
            seconds=t.seconds,
            microseconds=t.microseconds,
        )

    periods = [
        ("Y", d.years),
        ("M", d.months),
        ("D", d.remaining_days),
    ]
    period = "P"
    for sym, val in periods:
        period += f"{val}{sym}"
    times = [
        ("H", d.hours),
        ("M", d.minutes),
        ("S", d.remaining_seconds),
    ]
    time_ = "T"
    for sym, val in times:
        time_ += f"{val}{sym}"
    if d.microseconds:
        time_ = time_[:-1]
        time_ += f".{d.microseconds:06}S"
    return period + time_


@slotted(dict=False)
@dataclasses.dataclass(frozen=True)
class TaggedUnion:
    tag: str
    types: Tuple[Type, ...]
    isliteral: bool
    types_by_values: Tuple[Tuple[Any, Type], ...]


empty = object()


@functools.lru_cache(maxsize=None)
def get_tag_for_types(types: Tuple[Type, ...]) -> Optional[TaggedUnion]:
    if any(
        t in {None, ...} or not inspect.isclass(t) or checks.isstdlibtype(t)
        for t in types
    ):
        return None
    if len(types) > 1:
        root = types[0]
        root_hints = cached_type_hints(root)
        intersection = {*root_hints}
        fields_by_type = {root: root_hints}
        t: Type
        for t in types[1:]:
            hints = cached_type_hints(t)
            intersection &= hints.keys()
            fields_by_type[t] = hints
        tag = None
        literal = False
        # If we have an intersection, check if it's constant value we can use
        # TODO: This won't support Generics in this state.
        #  We don't support generics yet (#119), but when we do,
        #  we need to add a branch for tagged unions from generics.
        while intersection and tag is None:
            f = intersection.pop()
            v = getattr(root, f, empty)
            if v is not empty and not isinstance(v, MemberDescriptorType):
                tag = f
                continue
            rhint = root_hints[f]
            if checks.isliteral(rhint):
                tag, literal = f, True
        if tag:
            if literal:
                tbv = (
                    *((a, t) for t in types for a in get_args(fields_by_type[t][tag])),
                )
            else:
                tbv = (*((getattr(t, tag), t) for t in types),)
            return TaggedUnion(
                tag=tag, types=types, isliteral=literal, types_by_values=tbv
            )
    return None


def extract(name: str, *, frame: types.FrameType = None) -> Optional[Any]:
    """Extract `name` from the stacktrace of `frame`.

    If `frame` is not provided, this function will use the current frame.
    """
    frame = frame or inspect.currentframe()
    seen = set()
    while frame and frame not in seen:
        if name in frame.f_globals:
            return frame.f_globals[name]
        if name in frame.f_locals:
            return frame.f_locals[name]
        seen.add(frame)
        frame = frame.f_back

    return None



class ConstraintSyntaxError(SyntaxError):
    """A generic error indicating an improperly defined constraint."""

    pass


class ConstraintValueError(ValueError):
    """A generic error indicating a value violates a constraint."""

    pass



from __future__ import annotations

import dataclasses
from types import MappingProxyType
from typing import (
    Type,
    Tuple,
    cast,
    Mapping,
    Hashable,
    Any,
    Callable,
    Optional,
    Union,
)

from typic.api import wrap_cls, ObjectT
from typic.compat import DATACLASS_NATIVE_SLOTS, DATACLASS_KW_ONLY, DATACLASS_MATCH_ARGS
from typic.types import freeze
from typic.util import slotted
from .serde.common import SerdeFlags

_field_slots: Tuple[str, ...] = cast(Tuple[str, ...], dataclasses.Field.__slots__) + (
    "exclude",
    "external_name",
)

_EMPTY_METADATA: Mapping[Hashable, Any] = MappingProxyType({})
FactoryT = Callable[[], ObjectT]


class Field(dataclasses.Field):
    """An extension of :py:class:`dataclasses.Field` which adds a few parameters.

    This class adds some flags for determining the serialization protocol.

    Examples
    --------
    >>> import typic
    >>>
    >>> @typic.klass
    ... class Foo:
    ...     bar: str = typic.field(name="Bar")
    ...     exc: str = typic.field(exclude=True)
    ...
    >>> Foo("foo", "exc").primitive()
    {'Bar': 'foo'}
    """

    __slots__ = _field_slots

    def __init__(
        self,
        *,
        default: Union[ObjectT, dataclasses._MISSING_TYPE],
        default_factory: Union[FactoryT, dataclasses._MISSING_TYPE],
        init: bool,
        repr: Any,
        hash: Optional[Any],
        compare: bool,
        metadata: Optional[Mapping[Hashable, Any]],
        exclude: bool = False,
        name: str = None,
        **kwargs,
    ):
        super(Field, self).__init__(
            default, default_factory, init, repr, hash, compare, metadata, **kwargs  # type: ignore
        )
        self.exclude = exclude
        self.external_name = name

    @classmethod
    def from_field(cls: Type["Field"], f: dataclasses.Field) -> "Field":
        kwargs = dict(
            default=f.default,
            default_factory=f.default_factory,  # type: ignore
            init=f.init,
            repr=f.repr,
            hash=f.hash,
            compare=f.compare,
            metadata=f.metadata,  # type: ignore
        )
        if DATACLASS_KW_ONLY:
            kwargs["kw_only"] = f.kw_only  # type: ignore

        tf = cls(**kwargs)
        tf.name = f.name
        tf.type = f.type
        return tf


def field(
    *,
    default: Union[ObjectT, dataclasses._MISSING_TYPE] = dataclasses.MISSING,
    default_factory: Union[FactoryT, dataclasses._MISSING_TYPE] = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: Any = None,
    compare: bool = True,
    metadata: Mapping[Hashable, Any] = None,
    exclude: bool = False,
    name: str = None,
    kw_only: bool = False,
) -> Field:
    kwargs = dict(
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
        exclude=exclude,
        name=name,
    )
    if DATACLASS_KW_ONLY:
        kwargs["kw_only"] = kw_only
    return Field(**kwargs)  # type: ignore


def make_typedclass(
    cls: Type,
    *,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = False,
    unsafe_hash: bool = False,
    frozen: bool = False,
    delay: bool = False,
    strict: bool = False,
    jsonschema: bool = False,
    slots: bool = False,
    kw_only: bool = False,
    match_args: bool = True,
    serde: SerdeFlags = None,
):
    """A convenience function for generating a dataclass with type-coercion.

    Allows the user to create typed dataclasses on-demand from a base-class, i.e.::

        TypedClass = make_typedclass(UnTypedClass)

    The preferred method is via the `klass` decorator, however.

    See Also
    --------
    :py:func:`klass`
    :py:func:`dataclasses.dataclass`
    """
    # Make the base dataclass.
    kwargs = dict(
        init=init,
        repr=repr,
        eq=eq,
        order=order,
        unsafe_hash=unsafe_hash,
        frozen=frozen,
    )
    if DATACLASS_KW_ONLY:
        kwargs["kw_only"] = kw_only
    if DATACLASS_MATCH_ARGS:
        kwargs["match_args"] = match_args
    if DATACLASS_NATIVE_SLOTS:
        dcls = dataclasses.dataclass(cls, **kwargs, slots=slots)  # type: ignore
    else:
        dcls = dataclasses.dataclass(cls, **kwargs)  # type: ignore
        if slots:
            dcls = slotted(dcls)

    fields = [
        f if isinstance(f, Field) else Field.from_field(f)
        for f in dataclasses.fields(dcls)
    ]
    field_names = freeze({f.name: f.external_name or f.name for f in fields})
    exclude = frozenset({f.name for f in fields if f.exclude})
    dcls.__typic_fields__ = (*fields,)
    serde = serde or SerdeFlags()
    serde = serde.merge(SerdeFlags(fields=field_names, exclude=exclude))  # type: ignore
    return wrap_cls(
        dcls,
        delay=delay,
        strict=strict,
        jsonschema=jsonschema,
        serde=serde,
    )


def klass(
    _cls: Type = None,
    *,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = False,
    unsafe_hash: bool = False,
    frozen: bool = False,
    delay: bool = False,
    strict: bool = False,
    jsonschema: bool = True,
    slots: bool = False,
    kw_only: bool = False,
    match_args: bool = True,
    serde: SerdeFlags = None,
):
    """A convenience decorator for generating a dataclass with type-coercion.

    This::

        import typic

        @typic.klass
        class Foo:
            bar: str

    Is functionally equivalent to::

        import dataclasses
        import typic

        @typic.al
        @dataclasses.dataclass
        class Foo:
            bar: str

    See Also
    --------
    :py:func:`~typic.api.wrap_cls`
    :py:func:`dataclasses.dataclass`
    """

    def typedclass_wrapper(cls_):
        return make_typedclass(
            cls=cls_,
            init=init,
            repr=repr,
            eq=eq,
            order=order,
            unsafe_hash=unsafe_hash,
            frozen=frozen,
            delay=delay,
            strict=strict,
            jsonschema=jsonschema,
            slots=slots,
            kw_only=kw_only,
            match_args=match_args,
            serde=serde,
        )

    return typedclass_wrapper(_cls) if _cls else typedclass_wrapper
